# coding: utf-8
"""
Replace NOT NULL constraints per column by NOT NULL constraints conditioned by federation reference in objects_current.
"""

import flask_sqlalchemy


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    constraints = db.session.execute(db.text("""
         SELECT conname
         FROM pg_catalog.pg_constraint
         WHERE conname = 'users_not_null_check'
    """)).fetchall()

    if len(constraints) > 0:
        return False

    # Perform migration
    db.session.execute(db.text("""
        ALTER TABLE users
            ADD CONSTRAINT users_not_null_check
                CHECK (
                    (
                        type = 'FEDERATION_USER' AND
                        NOT is_admin AND
                        fed_id IS NOT NULL AND
                        component_id IS NOT NULL
                    ) OR (
                        name IS NOT NULL AND
                        email IS NOT NULL
                    )
                ),
            ALTER COLUMN name DROP NOT NULL,
            ALTER COLUMN email DROP NOT NULL
    """))
    return True



# coding: utf-8
"""
Models for users' favorite actions and instruments.
"""
import typing

from sqlalchemy.orm import Mapped, Query

from .. import db
from .utils import Model


class FavoriteAction(Model):
    __tablename__ = 'favorite_actions'

    action_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey("actions.id"), primary_key=True)
    user_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey("users.id"), primary_key=True)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["FavoriteAction"]]


class FavoriteInstrument(Model):
    __tablename__ = 'favorite_instruments'

    instrument_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey("instruments.id"), primary_key=True)
    user_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey("users.id"), primary_key=True)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["FavoriteInstrument"]]



"""
Add the result and expiration date column to the background_tasks table.
"""

import flask_sqlalchemy

from .utils import table_has_column


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    if table_has_column('background_tasks', 'result'):
        return False

    # Perform migration
    db.session.execute(db.text(
        """
        ALTER TABLE background_tasks
        ADD expiration_date TIMESTAMP NULL,
        ADD result JSON NULL
        """
    ))
    return True



# coding: utf-8
"""
Remove the default value for the declined column in the object location assignments table.
"""

import flask_sqlalchemy


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Perform migration
    db.session.execute(db.text("""
        ALTER TABLE object_location_assignments
        ALTER COLUMN declined DROP DEFAULT
    """))
    return True



# coding: utf-8
"""

"""

import datetime
import typing

from sqlalchemy.orm import relationship, Mapped, Query

from .. import db
from .objects import Objects
from .utils import Model

if typing.TYPE_CHECKING:
    from .users import User
    from .components import Component


class Comment(Model):
    __tablename__ = 'comments'
    __table_args__ = (
        db.CheckConstraint(
            '(fed_id IS NOT NULL AND component_id IS NOT NULL) OR (user_id IS NOT NULL AND utc_datetime IS NOT NULL)',
            name='comments_not_null_check'
        ),
        db.UniqueConstraint('fed_id', 'component_id', name='comments_fed_id_component_id_key')
    )

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    object_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(Objects.object_id_column), nullable=False)
    user_id: Mapped[typing.Optional[int]] = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=True)
    author: Mapped[typing.Optional['User']] = relationship('User')
    content: Mapped[str] = db.Column(db.Text, nullable=False)
    utc_datetime: Mapped[typing.Optional[datetime.datetime]] = db.Column(db.TIMESTAMP(timezone=True), nullable=True)
    fed_id: Mapped[typing.Optional[int]] = db.Column(db.Integer, nullable=True)
    component_id: Mapped[typing.Optional[int]] = db.Column(db.Integer, db.ForeignKey('components.id'), nullable=True)
    component: Mapped[typing.Optional['Component']] = relationship('Component')

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["Comment"]]

    def __init__(
            self,
            object_id: int,
            user_id: typing.Optional[int],
            content: str,
            utc_datetime: typing.Optional[datetime.datetime] = None,
            fed_id: typing.Optional[int] = None,
            component_id: typing.Optional[int] = None
    ) -> None:
        super().__init__(
            object_id=object_id,
            user_id=user_id,
            content=content,
            utc_datetime=utc_datetime if utc_datetime is not None else datetime.datetime.now(datetime.timezone.utc),
            fed_id=fed_id,
            component_id=component_id
        )

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(id={self.id}, object_id={self.object_id}, user_id={self.user_id}, utc_datetime={self.utc_datetime}, content="{self.content}")>'



# coding: utf-8
"""
Recreate the markdown_to_html_cache_entries table.
"""

import flask_sqlalchemy

from .utils import table_has_column
from ..markdown_to_html_cache import MarkdownToHTMLCacheEntry


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    if table_has_column('markdown_to_html_cache_entries', 'parameters'):
        return False

    # Perform migration
    db.session.execute(db.text("""
        DROP TABLE markdown_to_html_cache_entries
    """))
    db.session.commit()

    MarkdownToHTMLCacheEntry.__table__.create(bind=db.engine)  # type: ignore[attr-defined]

    return True



# coding: utf-8
"""

"""
import string
import typing

import flask
import flask_login
from flask_babel import _
from flask_wtf import FlaskForm
from wtforms import BooleanField, PasswordField, SelectField, SelectMultipleField
from wtforms.validators import InputRequired

from . import frontend
from .. import logic
from ..utils import object_permissions_required, FlaskResponseT
from ..models import Permissions, SciCatExportType


class SciCatExportForm(FlaskForm):
    tags = SelectMultipleField()
    owner_group = SelectField(validators=[InputRequired()])
    access_groups = SelectMultipleField()
    instrument = SelectField(validators=[InputRequired()])
    sample = SelectField(validators=[InputRequired()])
    input_datasets = SelectMultipleField()
    api_token = PasswordField()


class SciCatAPITokenForm(FlaskForm):
    api_token = PasswordField(validators=[InputRequired()])
    store_api_token = BooleanField()


@frontend.route('/objects/<int:object_id>/scicat_export/', methods=['GET', 'POST'])
@object_permissions_required(Permissions.GRANT)
def scicat_export(object_id: int) -> FlaskResponseT:
    existing_scicat_url = logic.scicat_export.get_scicat_url(object_id)
    if existing_scicat_url:
        return flask.render_template(
            'objects/scicat_export_already_exported.html',
            object_id=object_id,
            scicat_url=existing_scicat_url
        )

    api_url = flask.current_app.config['SCICAT_API_URL']
    frontend_url = flask.current_app.config['SCICAT_FRONTEND_URL']
    if not api_url or not frontend_url:
        return flask.abort(404)

    object = logic.objects.get_object(object_id)
    if object.eln_import_id is not None:
        return flask.abort(403)
    if object.component_id is not None or object.data is None or object.schema is None:
        flask.flash(_('Exporting imported objects is not supported.'), 'error')
        return flask.redirect(flask.url_for('.object', object_id=object_id))

    if object.action_id is not None:
        action = logic.actions.get_action(object.action_id)
        object_export_type = action.type.scicat_export_type if action.type is not None else None
    else:
        object_export_type = None
    if object_export_type is None:
        flask.flash(_("The SciCat export type is not set for objects of this type."), 'error')
        return flask.redirect(flask.url_for('.object', object_id=object_id))

    try:
        scicat_export_form = SciCatExportForm()

        user_id = flask_login.current_user.id
        api_token = logic.scicat_export.get_user_valid_api_token(api_url, user_id)
        had_invalid_api_token = False
        scicat_api_token_form = SciCatAPITokenForm()

        if not api_token:
            if scicat_api_token_form.validate_on_submit():
                form_api_token = scicat_api_token_form.api_token.data.strip()
                api_token_valid = logic.scicat_export.is_api_token_valid(api_url, form_api_token)
                if api_token_valid:
                    api_token = form_api_token
                    if scicat_api_token_form.store_api_token.data:
                        logic.settings.set_user_settings(user_id, {'SCICAT_API_TOKEN': api_token})
                    else:
                        scicat_export_form.api_token.data = api_token
                else:
                    had_invalid_api_token = True
                    api_token = None
            elif scicat_export_form.validate_on_submit():
                form_api_token = scicat_export_form.api_token.data.strip()
                api_token_valid = logic.scicat_export.is_api_token_valid(api_url, form_api_token)
                if api_token_valid:
                    api_token = form_api_token
                else:
                    had_invalid_api_token = True
                    api_token = None

        if not api_token:
            return flask.render_template(
                'objects/scicat_export_api_token.html',
                object_id=object_id,
                scicat_api_token_form=scicat_api_token_form,
                had_invalid_api_token=had_invalid_api_token
            )

        user_groups = logic.scicat_export.get_user_groups(api_url, api_token)
        if user_groups:
            scicat_export_form.owner_group.choices = [
                (group_name, group_name)
                for group_name in user_groups
            ]
            scicat_export_form.access_groups.choices = scicat_export_form.owner_group.choices[:]
        else:
            scicat_export_form.owner_group.choices = [('-', '—')]
            scicat_export_form.access_groups.choices = []

        scicat_export_form.instrument.choices = [('-', '—')]
        if object_export_type == SciCatExportType.RAW_DATASET:
            instruments = logic.scicat_export.get_instruments(api_url, api_token)
            if instruments:
                scicat_export_form.instrument.choices += instruments

        properties = {
            tuple(whitelist_path): metadata
            for metadata, whitelist_path, title_path in logic.dataverse_export.flatten_metadata(object.data)
            if not ('_type' in metadata and metadata['_type'] == 'tags') and all(c in string.ascii_letters + string.digits + '_' for c in ''.join(map(str, title_path)))
        }

        if object_export_type in {SciCatExportType.RAW_DATASET, SciCatExportType.DERIVED_DATASET}:
            tags_set = set()
            for property in object.data.values():
                if '_type' in property and property['_type'] == 'tags':
                    for tag in property['tags']:
                        tags_set.add(tag)
            tags = list(tags_set)
            tags.sort()

            scicat_export_form.tags.choices = [
                (tag, tag)
                for tag in tags
            ]
        else:
            tags = []
            scicat_export_form.tags.choices = []

        scicat_export_form.sample.choices = [('-', '—')]
        if object_export_type == SciCatExportType.RAW_DATASET:
            exported_referenced_objects = logic.scicat_export.get_exported_referenced_objects(object.data, object.schema, user_id)
            exported_referenced_samples = {
                export
                for export in exported_referenced_objects
                if export.type == SciCatExportType.SAMPLE
            }
            for export in exported_referenced_samples:
                if Permissions.READ in logic.object_permissions.get_user_object_permissions(object_id=export.object_id, user_id=user_id):
                    object_name = logic.objects.get_object(export.object_id).name
                    if object_name:
                        if isinstance(object_name, dict) and 'en' in object_name:
                            object_name = object_name['en']
                        else:
                            object_name = str(object_name)
                else:
                    object_name = None
                if not object_name:
                    object_name = f"Object #{export.object_id}"
                scicat_export_form.sample.choices.append((export.scicat_pid, object_name))

        scicat_export_form.input_datasets.choices = []
        if object_export_type == SciCatExportType.DERIVED_DATASET:
            exported_referenced_objects = logic.scicat_export.get_exported_referenced_objects(object.data, object.schema, user_id)
            exported_referenced_datasets = {
                export
                for export in exported_referenced_objects
                if export.type in {SciCatExportType.RAW_DATASET, SciCatExportType.DERIVED_DATASET}
            }
            for export in exported_referenced_datasets:
                if Permissions.READ in logic.object_permissions.get_user_object_permissions(object_id=export.object_id, user_id=user_id):
                    object_name = logic.objects.get_object(export.object_id).name
                    if object_name:
                        if isinstance(object_name, dict) and 'en' in object_name:
                            object_name = object_name['en']
                        else:
                            object_name = str(object_name)
                else:
                    object_name = None
                if not object_name:
                    object_name = f"Object #{export.object_id}"
                scicat_export_form.input_datasets.choices.append(('PID/' + export.scicat_pid, object_name))

        property_whitelist: typing.List[typing.List[typing.Union[str, int]]] = [
            ['name']
        ]
        for whitelist_path in properties:
            if 'property,' + ','.join(map(str, whitelist_path)) in flask.request.form:
                property_whitelist.append(list(whitelist_path))

        if 'do_export' in flask.request.form and scicat_export_form.validate_on_submit():
            tag_whitelist_set = set()
            for tag in scicat_export_form.tags.data:
                if tag in tags:
                    tag_whitelist_set.add(tag)
            tag_whitelist = list(tag_whitelist_set)
            tag_whitelist.sort()
            scicat_url = logic.scicat_export.upload_object(
                object_id=object_id,
                user_id=flask_login.current_user.id,
                api_url=api_url,
                frontend_url=frontend_url,
                api_token=api_token,
                property_whitelist=property_whitelist,
                tag_whitelist=tag_whitelist,
                object_export_type=object_export_type,
                owner_group=scicat_export_form.owner_group.data,
                access_groups=scicat_export_form.access_groups.data if scicat_export_form.access_groups.data else None,
                instrument_pid=scicat_export_form.instrument.data if scicat_export_form.instrument.data != '-' else None,
                sample_pid=scicat_export_form.sample.data if scicat_export_form.sample.data != '-' else None,
                input_dataset_pids=scicat_export_form.input_datasets.data if scicat_export_form.input_datasets.data else None
            )
            if scicat_url is not None:
                return flask.redirect(scicat_url)
            flask.flash(_("An unknown error occurred while uploading the dataset."), 'error')

        return flask.render_template(
            "objects/scicat_export.html",
            properties=properties,
            export_properties=property_whitelist,
            object_id=object_id,
            get_title_for_property=lambda path: logic.dataverse_export.get_title_for_property(path, object.schema),
            get_property_export_default=lambda path: logic.scicat_export.get_property_export_default(path, object.schema),
            scicat_export_form=scicat_export_form
        )
    except logic.errors.SciCatNotReachableError:
        flask.flash(_("%(scicat_name)s is currently unreachable, please try again later.", scicat_name=flask.current_app.config['SCICAT_NAME']), 'error')
        return flask.redirect(flask.url_for('.object', object_id=object_id))



# coding: utf-8
"""
RESTful API for SampleDB
"""
import json
import typing

import flask

from .authentication import multi_auth
from ..utils import Resource, ResponseData
from ...logic.actions import get_action, update_action
from ...logic.action_translations import set_action_translation
from ...logic.languages import Language
from ...logic.action_permissions import get_user_action_permissions, get_actions_with_permissions
from ...logic import errors, utils, actions, action_types
from ...logic.schemas.templates import find_invalid_template_paths
from ...logic.schemas.validate_schema import validate_schema
from ...models import Permissions

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


def action_to_json(action: actions.Action) -> typing.Dict[str, typing.Any]:
    return {
        'action_id': action.id,
        'instrument_id': action.instrument_id if not flask.current_app.config['DISABLE_INSTRUMENTS'] else None,
        'user_id': action.user_id,
        'type': {
            action_types.ActionType.SAMPLE_CREATION: 'sample',
            action_types.ActionType.MEASUREMENT: 'measurement',
            action_types.ActionType.SIMULATION: 'simulation'
        }.get(action.type_id, 'custom') if action.type_id is not None else 'custom',
        'type_id': action.type_id,
        'name': utils.get_translated_text(
            action.name,
            language_code='en'
        ) or None,
        'description': utils.get_translated_text(
            action.description,
            language_code='en'
        ) or None,
        'is_hidden': action.is_hidden,
        'schema': action.schema
    }


class Action(Resource):
    @multi_auth.login_required
    def get(self, action_id: int) -> ResponseData:
        try:
            action = get_action(
                action_id=action_id
            )
        except errors.ActionDoesNotExistError:
            return {
                "message": f"action {action_id} does not exist"
            }, 404
        if Permissions.READ not in get_user_action_permissions(action_id=action_id, user_id=flask.g.user.id):
            return flask.abort(403)
        return action_to_json(action)

    @multi_auth.login_required
    def post(self, action_id: int) -> ResponseData:
        try:
            action = get_action(
                action_id=action_id
            )
        except errors.ActionDoesNotExistError:
            return {
                "message": f"action {action_id} does not exist"
            }, 404
        if Permissions.WRITE not in get_user_action_permissions(action_id=action_id, user_id=flask.g.user.id):
            return flask.abort(403)
        if action.fed_id is not None:
            return flask.abort(403)
        action_json = action_to_json(action)
        request_json = flask.request.get_json(force=True)
        if not isinstance(request_json, dict):
            return {
                "message": "JSON object body required"
            }, 400
        name = action.name.get('en')
        description = action.description.get('en')
        short_description = action.short_description.get('en')
        schema = action.schema
        is_hidden = action.is_hidden
        for key in request_json:
            if key not in action_json:
                return {
                    "message": f"invalid key '{key}'"
                }, 400
            if isinstance(request_json[key], str) and '\0' in request_json[key]:
                return {
                    "message": f"{key} must not contain NULL"
                }, 400
            if key == 'name':
                if not isinstance(request_json['name'], str):
                    return {
                        "message": "name must be string"
                    }, 400
                if len(request_json['name']) < 1 or len(request_json['name']) > 100:
                    return {
                        "message": "name must be between 1 and 100 characters long"
                    }, 400
                name = request_json['name']
            elif key == 'description':
                if not isinstance(request_json['description'], str):
                    return {
                        "message": "description must be string"
                    }, 400
                description = request_json['description']
            elif key == 'is_hidden':
                if not isinstance(request_json['is_hidden'], bool):
                    return {
                        "message": "is_hidden must be boolean"
                    }, 400
                is_hidden = request_json['is_hidden']
            elif key == 'schema':
                if not isinstance(request_json['schema'], dict):
                    return {
                        "message": "schema must be dict"
                    }, 400
                schema = request_json['schema']
                error_message = None
                try:
                    invalid_template_paths = find_invalid_template_paths(schema, flask.g.user.id)
                    if invalid_template_paths:
                        raise errors.ValidationError('insufficient permissions for template action', invalid_template_paths[0])
                    validate_schema(schema, invalid_template_action_ids=[] if action is None else [action.id], strict=True)
                except errors.ValidationError as e:
                    error_message = e.message
                except Exception as e:
                    error_message = str(e)
                if error_message is not None:
                    return {
                        "message": error_message
                    }, 400
            else:
                if action_json[key] != request_json[key]:
                    return {
                        "message": f"{key} must be {json.dumps(action_json[key])}"
                    }, 400
        update_action(
            action_id=action_id,
            schema=schema,
            description_is_markdown=action.description_is_markdown,
            is_hidden=is_hidden,
            short_description_is_markdown=action.short_description_is_markdown,
        )
        set_action_translation(
            language_id=Language.ENGLISH,
            action_id=action_id,
            name=name or '',
            description=description or '',
            short_description=short_description or ''
        )
        action = get_action(action_id=action_id)
        return action_to_json(action)


class Actions(Resource):
    @multi_auth.login_required
    def get(self) -> ResponseData:
        actions = get_actions_with_permissions(user_id=flask.g.user.id, permissions=Permissions.READ)
        return [
            action_to_json(action)
            for action in actions
        ]



# coding: utf-8
"""
Basic configuration for SampleDB

This configuration is the pure base, representing defaults. These values may be altered or expanded in several ways:
- For tests, the configuration is modified in tests/conftest.py.
- For local, interactive testing and demonstrations, the configuration is modified in demo.py.
- environment variables starting with the prefix SAMPLEDB_ will further override any hardcoded configuration data.
"""

import base64
import io
import json
import os
import sys
import typing

import pytz
import pytz.exceptions
import requests
import sqlalchemy
from PIL import Image

from .utils import generate_secret_key, load_environment_configuration, ansi_color, text_to_bool
from .frontend.labels import PAGE_SIZE_KEYS, PAGE_SIZES
from .models.notifications import NotificationMode, NotificationType

REQUIRED_CONFIG_KEYS: typing.Set[str] = {
    'SQLALCHEMY_DATABASE_URI',
    'MAIL_SERVER',
    'MAIL_SENDER',
    'CONTACT_EMAIL',
}

LDAP_REQUIRED_CONFIG_KEYS: typing.Set[str] = {
    'LDAP_NAME',
    'LDAP_SERVER',
    'LDAP_USER_BASE_DN',
    'LDAP_UID_FILTER',
    'LDAP_NAME_ATTRIBUTE',
    'LDAP_MAIL_ATTRIBUTE',
    'LDAP_OBJECT_DEF',
}

OIDC_REQUIRED_CONFIG_KEYS: typing.Set[str] = {
    'OIDC_NAME',
    'OIDC_ISSUER',
    'OIDC_CLIENT_ID',
    'OIDC_CLIENT_SECRET',
}


def use_environment_configuration(env_prefix: str) -> None:
    """
    Uses configuration data from environment variables with a given prefix by setting the config modules variables.
    """
    config = load_environment_configuration(env_prefix)
    for name, value in config.items():
        globals()[name] = value
    parse_configuration_values()


def parse_configuration_values() -> None:
    """
    Convert configuration values from strings to integers, booleans or JSON objects.
    """
    # parse values as integers
    for config_name in [
        'INVITATION_TIME_LIMIT',
        'MAX_CONTENT_LENGTH',
        'MAX_BATCH_SIZE',
        'VALID_TIME_DELTA',
        'DOWNLOAD_SERVICE_TIME_LIMIT',
        'TYPEAHEAD_OBJECT_LIMIT',
        'LDAP_CONNECT_TIMEOUT',
        'TEMPORARY_FILE_TIME_LIMIT',
        'SHARED_DEVICE_SIGN_OUT_MINUTES',
        'MIN_NUM_TEXT_CHOICES_FOR_SEARCH',
        'PDFEXPORT_LOGO_WIDTH',
    ]:
        value = globals().get(config_name)
        if isinstance(value, str):
            try:
                globals()[config_name] = int(value)
            except Exception:
                pass

    # parse values as json
    for config_name in [
        'SERVICE_DESCRIPTION',
        'EXTRA_USER_FIELDS',
        'DOWNLOAD_SERVICE_WHITELIST',
        'LABEL_PAPER_FORMATS',
        'DEFAULT_NOTIFICATION_MODES'
    ]:
        value = globals().get(config_name)
        if isinstance(value, str) and (value.startswith('{') or value.startswith('[')):
            try:
                globals()[config_name] = json.loads(value)
            except Exception:
                pass

    # parse boolean values
    for config_name in [
        'DISABLE_USER_INVITATIONS',
        'ONLY_ADMINS_CAN_MANAGE_LOCATIONS',
        'ONLY_ADMINS_CAN_CREATE_GROUPS',
        'ONLY_ADMINS_CAN_DELETE_GROUPS',
        'ONLY_ADMINS_CAN_CREATE_PROJECTS',
        'ONLY_ADMINS_CAN_DELETE_PROJECTS',
        'ONLY_ADMINS_CAN_MANAGE_GROUP_CATEGORIES',
        'DISABLE_USE_IN_MEASUREMENT',
        'DISABLE_SUBPROJECTS',
        'ENFORCE_SPLIT_NAMES',
        'BUILD_TRANSLATIONS',
        'SHOW_PREVIEW_WARNING',
        'SHOW_OBJECT_TITLE',
        'FULL_WIDTH_OBJECTS_TABLE',
        'HIDE_OBJECT_TYPE_AND_ID_ON_OBJECT_PAGE',
        'DISABLE_INLINE_EDIT',
        'ENABLE_BACKGROUND_TASKS',
        'ENABLE_MONITORINGDASHBOARD',
        'ENABLE_ANONYMOUS_USERS',
        'ENABLE_NUMERIC_TAGS',
        'SHOW_UNHANDLED_OBJECT_RESPONSIBILITY_ASSIGNMENTS',
        'SHOW_LAST_PROFILE_UPDATE',
        'USE_TYPEAHEAD_FOR_OBJECTS',
        'DISABLE_INSTRUMENTS',
        'ENABLE_FUNCTION_CACHES',
        'ENABLE_CONTENT_SECURITY_POLICY',
        'ENABLE_ELN_FILE_IMPORT',
        'ELN_FILE_IMPORT_ALLOW_HTTP',
        'ENABLE_FEDERATION_DISCOVERABILITY',
        'ENABLE_WEBHOOKS_FOR_USERS',
        'WEBHOOKS_ALLOW_HTTP',
        'ENABLE_FIDO2_PASSKEY_AUTHENTICATION',
        'DISABLE_OUTDATED_USE_AS_TEMPLATE',
        'DISABLE_TOPICS',
        'ENABLE_FEDERATED_LOGIN',
        'ENABLE_FEDERATED_LOGIN_CREATE_NEW_USER',
        'OIDC_ONLY',
        'OIDC_DISABLE_NONCE',
        'OIDC_ACCESS_TOKEN_AS_API_KEY',
        'OIDC_ACCESS_TOKEN_ALLOW_INTROSPECTION',
        'WORKFLOW_VIEW_MODALS',
        'WORKFLOW_VIEW_COLLAPSED',
    ]:
        value = globals().get(config_name)
        if isinstance(value, str):
            globals()[config_name] = text_to_bool(value)


def is_download_service_whitelist_valid() -> bool:
    # check if paths and user ids from DOWNLOAD_SERVICE_WHITELIST are valid
    for path, user_ids in DOWNLOAD_SERVICE_WHITELIST.items():
        norm_path = os.path.normpath(os.path.join(os.path.sep, path))
        if norm_path != path:
            print(
                ansi_color(
                    'DOWNLOAD_SERVICE_WHITELIST: Please use a normalized '
                    'paths.'
                    '\n',
                    color=33
                ),
                file=sys.stderr
            )
            return False

        for user_id in user_ids:
            if type(user_id) is not int:
                print(
                    ansi_color(
                        'DOWNLOAD_SERVICE_WHITELIST: Please use a number for '
                        'the user IDs.'
                        '\n',
                        color=33
                    ),
                    file=sys.stderr
                )
                return False
    return True


def is_label_paper_formats_valid() -> bool:
    """check if paper formats from LABEL_PAPER_FORMATS are valid"""

    if not isinstance(LABEL_PAPER_FORMATS, list):
        print(ansi_color('LABEL_PAPER_FORMATS: Must be a list of dictionaries.\n', color=31))
        return False
    is_valid = True
    for i, format_definition in enumerate(LABEL_PAPER_FORMATS, start=1):
        if not isinstance(format_definition, dict):
            print(ansi_color('LABEL_PAPER_FORMATS: Must be a list of dictionaries.\n', color=31))
            return False
        str_fields = ['format_name']
        int_fields = ['labels_in_row', 'labels_in_col', 'qr_code_width', 'paper_format']
        float_fields = ['label_width', 'label_height', 'margin_horizontal', 'margin_vertical']
        for key, value in format_definition.items():
            if key in int_fields:
                int_fields.remove(key)
                if not isinstance(value, int):
                    is_valid = False
                    print(ansi_color(f'LABEL_PAPER_FORMATS: {key} must be a whole number in format definition {i}.\n', color=31))
                if key == 'qr_code_width':
                    if value < 4:
                        format_definition['qr_code_width'] = 4
                        print(ansi_color(f'LABEL_PAPER_FORMATS: {key} in format definition {i} is less than 4. Value set to 4.\n', color=33))
                    elif value > 150:
                        format_definition['qr_code_width'] = 150
                        print(ansi_color(f'LABEL_PAPER_FORMATS: {key} in format definition {i} is greater than 150. Value set to 150.\n', color=33))
                elif value < 0:
                    is_valid = False
                    print(ansi_color(f'LABEL_PAPER_FORMATS: {key} must be greater than 0 in format definition {i}.\n', color=31))
            elif key in float_fields:
                float_fields.remove(key)
                if not isinstance(value, (int, float)):
                    is_valid = False
                    print(ansi_color(f'LABEL_PAPER_FORMATS: {key} must be a number in format definition {i}.\n', color=31))
                if value < 0:
                    is_valid = False
                    print(ansi_color(f'LABEL_PAPER_FORMATS: {key} must be greater than 0 in format definition {i}.\n', color=31))
            elif key in str_fields:
                str_fields.remove(key)
                if not isinstance(value, str) and not isinstance(value, dict):
                    is_valid = False
                    print(ansi_color(f'LABEL_PAPER_FORMATS: {key} must be a string in format definition {i}.\n', color=31))
            else:
                print(ansi_color(f'LABEL_PAPER_FORMATS: {key} is an Unknown key in format definition {i}.\n', color=33))
        if not is_valid:
            return False

        if len(str_fields) > 0 or len(int_fields) > 0 or len(float_fields) > 0:
            for key in str_fields + int_fields + float_fields:
                print(ansi_color(f'LABEL_PAPER_FORMATS: {key} is missing in format definition {i}.\n', color=31))
            return False

        if format_definition['paper_format'] < 0 or format_definition['paper_format'] >= len(PAGE_SIZE_KEYS):
            print(ansi_color(f'LABEL_PAPER_FORMATS: invalid paper_format in format definition {i}.\n', color=31))
            return False
        page_size = PAGE_SIZES[PAGE_SIZE_KEYS[format_definition['paper_format']]]
        page_width = page_size[0]
        page_height = page_size[1]
        used_page_width = format_definition['labels_in_row'] * format_definition['label_width'] + (format_definition['labels_in_row'] - 1) * format_definition['margin_horizontal']
        used_page_height = format_definition['labels_in_col'] * format_definition['label_height'] + (format_definition['labels_in_col'] - 1) * format_definition['margin_vertical']
        if page_width < used_page_width or page_height < used_page_height:
            print(ansi_color(f'LABEL_PAPER_FORMATS: page is too small for labels in format definition {i}.\n', color=31))
            return False

    return is_valid


def is_default_notification_modes_valid() -> bool:
    """
    Check if the default notification mode dict is valid.
    """
    if DEFAULT_NOTIFICATION_MODES is None:
        return True
    if not isinstance(DEFAULT_NOTIFICATION_MODES, dict):
        print(ansi_color(f'DEFAULT_NOTIFICATION_MODES must be dict, but got {type(DEFAULT_NOTIFICATION_MODES)}.\n', color=31))
        return False
    for key, value in DEFAULT_NOTIFICATION_MODES.items():
        valid_keys = [notification_type.name.upper() for notification_type in NotificationType] + ["DEFAULT"]
        if key not in valid_keys:
            print(ansi_color(f'DEFAULT_NOTIFICATION_MODES keys must be one of {valid_keys!r}, but got {key!r}.\n', color=31))
            return False
        valid_modes = [notification_mode.name.upper() for notification_mode in NotificationMode]
        if value not in valid_modes:
            print(ansi_color(f'DEFAULT_NOTIFICATION_MODES values must be one of {valid_modes!r}, but got {value!r}.\n', color=31))
            return False
    return True


def check_config(
        config: typing.Mapping[str, typing.Any]
) -> typing.Dict[str, typing.Any]:
    """
    Check whether all neccessary configuration values are set.

    Print a warning if missing values will lead to reduced functionality.
    Exit if missing values will prevent SampleDB from working correctly.

    :param config: the config mapping
    """
    defined_config_keys = {
        key
        for key, value in config.items()
        if value is not None
    }

    show_config_info = False
    can_run = True

    internal_config: typing.Dict[str, typing.Any] = {}

    missing_config_keys = REQUIRED_CONFIG_KEYS - defined_config_keys

    if missing_config_keys:
        print(
            ansi_color(
                'Missing required configuration values:\n - ' +
                '\n - '.join(missing_config_keys) +
                '\n',
                color=31
            ),
            file=sys.stderr
        )
        can_run = False
        show_config_info = True

    missing_config_keys = LDAP_REQUIRED_CONFIG_KEYS - defined_config_keys
    if missing_config_keys:
        print(
            'LDAP authentication will be disabled, because the following '
            'configuration values are missing:\n -',
            '\n - '.join(missing_config_keys),
            '\n',
            file=sys.stderr
        )
        show_config_info = True

    missing_config_keys = OIDC_REQUIRED_CONFIG_KEYS - defined_config_keys
    if missing_config_keys:
        print(
            'OIDC authentication will be disabled, because the following '
            'configuration values are missing:\n -',
            '\n - '.join(missing_config_keys),
            '\n',
            file=sys.stderr
        )
        show_config_info = True

    if 'JUPYTERHUB_URL' not in defined_config_keys and 'JUPYTERHUB_TEMPLATES_URL' not in defined_config_keys:
        print(
            'JupyterHub integration will be disabled, because none of following '
            'configuration values are defined:\n -',
            '\n - '.join(['JUPYTERHUB_URL', 'JUPYTERHUB_TEMPLATES_URL']),
            '\n',
            file=sys.stderr
        )
        show_config_info = True

    if 'DATAVERSE_URL' not in defined_config_keys:
        print(
            'Dataverse export will be disabled, because the configuration '
            'value DATAVERSE_URL is missing.\n'
            '\n',
            file=sys.stderr
        )
        show_config_info = True

    if 'SCICAT_API_URL' not in defined_config_keys:
        print(
            'SciCat export will be disabled, because the configuration '
            'value SCICAT_API_URL is missing.\n'
            '\n',
            file=sys.stderr
        )
        show_config_info = True
    elif 'SCICAT_FRONTEND_URL' not in defined_config_keys:
        print(
            'SciCat export will be disabled, because the configuration '
            'value SCICAT_FRONTEND_URL is missing.\n'
            '\n',
            file=sys.stderr
        )
        show_config_info = True

    if 'DOWNLOAD_SERVICE_URL' not in defined_config_keys:
        print(
            'Download service will be disabled, because the configuration '
            'value DOWNLOAD_SERVICE_URL is missing.\n'
            '\n',
            file=sys.stderr
        )
        show_config_info = True
    elif 'DOWNLOAD_SERVICE_SECRET' not in defined_config_keys:
        print(
            'Download service will be disabled, because the configuration '
            'value DOWNLOAD_SERVICE_SECRET is missing.\n'
            '\n',
            file=sys.stderr
        )
        show_config_info = True

    if 'ENABLE_FEDERATED_LOGIN' not in defined_config_keys and 'ENABLE_FEDERATED_LOGIN_CREATE_NEW_USER' in defined_config_keys:
        print(
            'The configuration value ENABLE_FEDERATED_LOGIN_CREATE_NEW_USER '
            'will have no effect, because the configuration value '
            'ENABLE_FEDERATED_LOGIN is missing.\n'
            '\n'
        )
        show_config_info = True

    admin_password_set = 'ADMIN_PASSWORD' in defined_config_keys
    admin_username_set = 'ADMIN_USERNAME' in defined_config_keys
    admin_email_set = 'ADMIN_EMAIL' in defined_config_keys
    if admin_password_set or admin_username_set or admin_email_set:
        if not admin_password_set:
            if admin_username_set and admin_email_set:
                print(
                    'ADMIN_USERNAME and ADMIN_EMAIL are set, but '
                    'ADMIN_PASSWORD is missing. No admin user will be created.'
                    '\n',
                    file=sys.stderr
                )
            elif admin_username_set:
                print(
                    'ADMIN_USERNAME is set, but ADMIN_PASSWORD is missing. No '
                    'admin user will be created.'
                    '\n',
                    file=sys.stderr
                )
            elif admin_email_set:
                print(
                    'ADMIN_EMAIL is set, but ADMIN_PASSWORD is missing. No '
                    'admin user will be created.'
                    '\n',
                    file=sys.stderr
                )
        elif config['ADMIN_PASSWORD'] == '':
            print(
                'ADMIN_PASSWORD is an empty string. No admin user will be '
                'created.'
                '\n',
                file=sys.stderr
            )
        elif len(config['ADMIN_PASSWORD']) < 8:
            print(
                'ADMIN_PASSWORD is too short. No admin user will be created.'
                '\n',
                file=sys.stderr
            )
        elif can_run:
            engine = sqlalchemy.create_engine(config['SQLALCHEMY_DATABASE_URI'], **config['SQLALCHEMY_ENGINE_OPTIONS'])
            with engine.begin() as connection:
                user_table_exists = bool(connection.execute(sqlalchemy.text(
                    "SELECT * "
                    "FROM information_schema.columns "
                    "WHERE table_name = 'users'"
                )).fetchall())
                if user_table_exists:
                    users_exist = bool(connection.execute(sqlalchemy.text(
                        "SELECT * FROM users"
                    )).fetchall())
                else:
                    users_exist = False
            if users_exist:
                print(
                    'ADMIN_PASSWORD is set, but there already are users in '
                    'the database. No admin user will be created.'
                    '\n',
                    file=sys.stderr
                )
            else:
                admin_username = config.get('ADMIN_USERNAME', 'admin').lower()
                admin_email = config.get('ADMIN_EMAIL', config['CONTACT_EMAIL']).lower()
                print(
                    f'A new admin user with the username "{admin_username}", the email '
                    f'address "{admin_email}" and the given ADMIN_PASSWORD will be '
                    'created.'
                    '\n',
                    file=sys.stderr
                )
                internal_config['ADMIN_INFO'] = (
                    admin_username, admin_email, config['ADMIN_PASSWORD']
                )
                if config['ADMIN_PASSWORD'] == 'password':
                    print(
                        ansi_color(
                            'You are using the default ADMIN_PASSWORD from the '
                            'SampleDB documentation. Please sign in and change your '
                            'password before making this SampleDB instance available '
                            'to other users.'
                            '\n',
                            color=33
                        ),
                        file=sys.stderr
                    )

        show_config_info = True

    if can_run:
        engine = sqlalchemy.create_engine(config['SQLALCHEMY_DATABASE_URI'], **config['SQLALCHEMY_ENGINE_OPTIONS'])
        with engine.begin() as connection:
            files_table_with_data_exists = bool(connection.execute(sqlalchemy.text(
                """
                SELECT COUNT(*)
                FROM information_schema.columns
                WHERE table_name = 'files' AND column_name = 'data'
                """
            )).scalar())
            if files_table_with_data_exists:
                local_files_exist = bool(connection.execute(sqlalchemy.text(
                    """
                    SELECT COUNT(*)
                    FROM files
                    WHERE data->>'storage' = 'local'
                    """
                )).scalar())
            else:
                # if there is no file table with a data column, local files may
                # still exist from a version of SampleDB before the data column
                # was added
                local_files_exist = bool(connection.execute(sqlalchemy.text(
                    """
                    SELECT COUNT(*)
                    FROM information_schema.columns
                    WHERE table_name = 'files' AND column_name = 'original_file_name'
                    """
                )).scalar())
            if local_files_exist:
                print(
                    ansi_color(
                        'There are files with "local" storage. These are not '
                        'supported anymore starting with SampleDB 0.26.0. '
                        'Please move them to "database" storage using an '
                        'earlier version of SampleDB. For more information, '
                        'see: '
                        'https://scientific-it-systems.iffgit.fz-juelich.de/SampleDB/administrator_guide/deprecated_features.html#local-file-storage'
                        '\n',
                        color=31
                    ),
                    file=sys.stderr
                )
                can_run = False
                show_config_info = True
            elif 'FILE_STORAGE_PATH' in defined_config_keys:
                print(
                    ansi_color(
                        'FILE_STORAGE_PATH is set, however files with "local" storage '
                        'are not supported anymore starting with SampleDB 0.26.0. '
                        'Please unset the FILE_STORAGE_PATH config variable and do '
                        'not mount a directory for local file storage. For more '
                        'information, see: '
                        'https://scientific-it-systems.iffgit.fz-juelich.de/SampleDB/administrator_guide/deprecated_features.html#local-file-storage'
                        '\n',
                        color=31
                    ),
                    file=sys.stderr
                )
                can_run = False
                show_config_info = True

    internal_config['PDFEXPORT_LOGO_WIDTH'] = min(200, max(0, config['PDFEXPORT_LOGO_WIDTH']))
    if config['PDFEXPORT_LOGO_URL'] is not None:
        logo_url = config['PDFEXPORT_LOGO_URL']
        logo_image: typing.Optional[Image.Image] = None
        if logo_url.startswith('file://'):
            logo_path = logo_url[7:]
            try:
                logo_path = os.path.abspath(logo_path)
                _, logo_extension = os.path.splitext(logo_path)
                if logo_extension.lower() in ('.png', '.jpg', '.jpeg'):
                    logo_image = Image.open(logo_path)
                else:
                    print(
                        ansi_color(
                            f'Unsupported logo file format: {logo_extension}\n',
                            color=33
                        ),
                        file=sys.stderr
                    )
            except Exception:
                print(
                    ansi_color(
                        f'Unable to read logo file at: {logo_path}\n',
                        color=33
                    ),
                    file=sys.stderr
                )
        elif logo_url.startswith('http://') or logo_url.startswith('https://'):
            try:
                r = requests.get(logo_url, timeout=5)
                if r.status_code != 200:
                    print(
                        ansi_color(
                            f'Unable to read logo from: {logo_url}. Got status code: {r.status_code}\n',
                            color=33
                        ),
                        file=sys.stderr
                    )
                else:
                    logo_file = io.BytesIO(r.content)
                    logo_image = Image.open(logo_file)
            except Exception:
                print(
                    ansi_color(
                        f'Unable to read logo from: {logo_url}\n',
                        color=33
                    ),
                    file=sys.stderr
                )
        else:
            print(
                ansi_color(
                    f'Unable to read logo from: {logo_url}. The following URL schemes are supported: file, http, https.\n',
                    color=33
                ),
                file=sys.stderr
            )
        if logo_image:
            try:
                logo_width, logo_height = logo_image.size
                internal_config['PDFEXPORT_LOGO_ASPECT_RATIO'] = logo_width / logo_height
                logo_image = logo_image.convert('RGBA')
                background_image = Image.new('RGBA', logo_image.size, 'white')
                logo_image = Image.alpha_composite(background_image, logo_image)
                logo_file = io.BytesIO()
                logo_image.save(logo_file, "png")
                logo_png_data = logo_file.getvalue()
                logo_data_uri = 'data:image/png;base64,' + base64.b64encode(logo_png_data).decode('utf-8')
                internal_config['PDFEXPORT_LOGO_URL'] = logo_data_uri
            except Exception:
                print(
                    ansi_color(
                        f'Unable to read logo from: {logo_url}\n',
                        color=33
                    ),
                    file=sys.stderr
                )

    if config['TIMEZONE']:
        try:
            pytz.timezone(config['TIMEZONE'])
        except pytz.exceptions.UnknownTimeZoneError:
            print(
                ansi_color(
                    'Unknown time zone.\n',
                    color=31
                ),
                file=sys.stderr
            )
            can_run = False
            show_config_info = True

    if not isinstance(config['INVITATION_TIME_LIMIT'], int) or config['INVITATION_TIME_LIMIT'] <= 0:
        print(
            ansi_color(
                f'Expected INVITATION_TIME_LIMIT to be a positive integer, but got {config["INVITATION_TIME_LIMIT"]!r}\n',
                color=31
            ),
            file=sys.stderr
        )
        can_run = False
        show_config_info = True

    if not isinstance(config['EXTRA_USER_FIELDS'], dict):
        print(
            ansi_color(
                f'Expected EXTRA_USER_FIELDS to be a dictionary, but got {config["EXTRA_USER_FIELDS"]!r}\n',
                color=31
            ),
            file=sys.stderr
        )
        can_run = False
        show_config_info = True
    elif any(
            not isinstance(value, dict) or
            set(value.keys()) - {'name', 'placeholder'} or
            not isinstance(value.get('name', ''), (dict, str)) or
            not isinstance(value.get('placeholder', ''), (dict, str))
            for key, value in config['EXTRA_USER_FIELDS'].items()
    ):
        print(
            ansi_color(
                'Invalid EXTRA_USER_FIELDS.\n',
                color=31
            ),
            file=sys.stderr
        )
        can_run = False
        show_config_info = True

    if not is_download_service_whitelist_valid():
        can_run = False
        show_config_info = True

    if not is_label_paper_formats_valid():
        can_run = False
        show_config_info = True

    if not is_default_notification_modes_valid():
        can_run = False
        show_config_info = True

    if config['OIDC_CREATE_ACCOUNT'] not in ('no', 'deny_existing', 'auto_link'):
        can_run = False
        show_config_info = True

    if show_config_info:
        print(
            'For more information on setting SampleDB configuration, see: '
            'https://scientific-it-systems.iffgit.fz-juelich.de/SampleDB/'
            'administrator_guide/configuration.html',
            file=sys.stderr
        )

    if not can_run:
        sys.exit(1)

    return internal_config


# prefix for all routes (used by run script)
SERVER_PATH = '/'

# whether to use CSRF protection
# see: https://flask-wtf.readthedocs.io/en/stable/config.html
CSRF_ENABLED = True

# secret key for Flask, wtforms and more
# see: http://flask.pocoo.org/docs/1.0/config/#SECRET_KEY
# automatically generated default, but should be replaced using environment variable SAMPLEDB_SECRET_KEY
SECRET_KEY = generate_secret_key(num_bits=256)

# sameSite attribute for cookies
# see: https://flask.palletsprojects.com/en/2.2.x/security/#set-cookie-options
SESSION_COOKIE_SAMESITE = 'Lax'

# SQLAlchemy settings
SQLALCHEMY_ENGINE_OPTIONS = {'connect_args': {"options": "-c timezone=utc"}}

# LDAP settings
LDAP_NAME = None
LDAP_SERVER = None
LDAP_USER_BASE_DN = None
LDAP_UID_FILTER = None
LDAP_NAME_ATTRIBUTE = None
LDAP_MAIL_ATTRIBUTE = None
LDAP_OBJECT_DEF = None
LDAP_CONNECT_TIMEOUT = 5
# LDAP credentials, may both be None if anonymous access is enabled
LDAP_USER_DN = None
LDAP_PASSWORD = None

# OIDC settings
OIDC_NAME = None
OIDC_ISSUER = None
OIDC_CLIENT_ID = None
OIDC_CLIENT_SECRET = None
OIDC_SCOPES = 'openid profile email'
OIDC_DISABLE_NONCE = False
OIDC_ROLES = None
OIDC_ONLY = False
OIDC_CREATE_ACCOUNT = 'auto_link'
OIDC_ACCESS_TOKEN_AS_API_KEY = False
OIDC_ACCESS_TOKEN_ALLOW_INTROSPECTION = False

# email settings
MAIL_SERVER = None
MAIL_SENDER = None
MAIL_REPLY_TO = None
CONTACT_EMAIL = None

# branding and legal info
SERVICE_NAME = 'SampleDB'
SERVICE_DESCRIPTION = {
    'en': SERVICE_NAME + ' is a database for sample and measurement metadata developed at PGI and JCNS.',
    'de': SERVICE_NAME + ' ist eine Datenbank für Proben- und Messungsmetadaten entwickelt am PGI und JCNS.'
}
SERVICE_IMPRINT = None
SERVICE_LEGAL_NOTICE = None
SERVICE_PRIVACY_POLICY = None
SERVICE_ACCESSIBILITY = None
HELP_URL = 'https://scientific-it-systems.iffgit.fz-juelich.de/SampleDB/#documentation'

# a map of file extensions and the MIME types they should be handled as
# this is used to determine which user uploaded files should be served as
# images (those with an image/ MIME type) in the object view and which support
# a preview.
# files with extensions not listed here only support being downloaded.
MIME_TYPES = {
    '.txt': 'text/plain',
    '.png': 'image/png',
    '.jpg': 'image/jpeg',
    '.gif': 'image/gif',
    '.pdf': 'application/pdf'
}

# JupyterHub settings
JUPYTERHUB_NAME = 'JupyterHub'
JUPYTERHUB_URL = None
JUPYTERHUB_TEMPLATES_URL = None

# Dataverse settings
DATAVERSE_NAME = 'Dataverse'
DATAVERSE_URL = None
DATAVERSE_ROOT_IDS = ':root'

# Scicat settings
SCICAT_NAME = 'SciCat'
SCICAT_API_URL = None
SCICAT_FRONTEND_URL = None
SCICAT_EXTRA_PID_PREFIX = ''

DOWNLOAD_SERVICE_URL = None
DOWNLOAD_SERVICE_SECRET = None
DOWNLOAD_SERVICE_WHITELIST: typing.Dict[str, typing.List[int]] = {}
DOWNLOAD_SERVICE_TIME_LIMIT = 24 * 60 * 60

# PDF export settings
PDFEXPORT_LOGO_URL = None
PDFEXPORT_LOGO_ALIGNMENT = 'right'
PDFEXPORT_LOGO_WIDTH = 30

# label settings
LABEL_PAPER_FORMATS: list[dict[str, typing.Any]] = []

# CSRF token time limit
# users may take a long time to fill out a form during an experiment
WTF_CSRF_TIME_LIMIT = 12 * 60 * 60

# invitation link time limit
INVITATION_TIME_LIMIT = 7 * 24 * 60 * 60
DISABLE_USER_INVITATIONS = False

# Flask-MonitoringDashboard settings
ENABLE_MONITORINGDASHBOARD = False
MONITORINGDASHBOARD_DATABASE = 'sqlite:///flask_monitoringdashboard.db'

# other settings
ONLY_ADMINS_CAN_MANAGE_LOCATIONS = False
ONLY_ADMINS_CAN_CREATE_GROUPS = False
ONLY_ADMINS_CAN_DELETE_GROUPS = False
ONLY_ADMINS_CAN_CREATE_PROJECTS = False
ONLY_ADMINS_CAN_DELETE_PROJECTS = False
ONLY_ADMINS_CAN_MANAGE_GROUP_CATEGORIES = True

DISABLE_USE_IN_MEASUREMENT = False

DISABLE_SUBPROJECTS = False

USE_TYPEAHEAD_FOR_OBJECTS = False
TYPEAHEAD_OBJECT_LIMIT = None

ENFORCE_SPLIT_NAMES = False

BUILD_TRANSLATIONS = True
PYBABEL_PATH = 'pybabel'

EXTRA_USER_FIELDS: typing.Dict[str, typing.Dict[str, typing.Dict[str, str]]] = {}

SHOW_PREVIEW_WARNING = False

DISABLE_INLINE_EDIT = False

SHOW_OBJECT_TITLE = False
WORKFLOW_VIEW_MODALS = False
WORKFLOW_VIEW_COLLAPSED = False

FULL_WIDTH_OBJECTS_TABLE = True

HIDE_OBJECT_TYPE_AND_ID_ON_OBJECT_PAGE = False

MAX_BATCH_SIZE = 100

FEDERATION_UUID = None
ENABLE_FEDERATION_DISCOVERABILITY = True
ALLOW_HTTP = False
VALID_TIME_DELTA = 300
ENABLE_DEFAULT_USER_ALIASES = False
ENABLE_FEDERATED_LOGIN = False
ENABLE_FEDERATED_LOGIN_CREATE_NEW_USER = False

ENABLE_WEBHOOKS_FOR_USERS = False
WEBHOOKS_ALLOW_HTTP = False

ENABLE_BACKGROUND_TASKS = True

TIMEZONE = None

ENABLE_ANONYMOUS_USERS = False

ENABLE_NUMERIC_TAGS = None

SHOW_UNHANDLED_OBJECT_RESPONSIBILITY_ASSIGNMENTS = True

SHOW_LAST_PROFILE_UPDATE = True

DISABLE_INSTRUMENTS = False

ENABLE_FUNCTION_CACHES = True

# temporary file time limit
TEMPORARY_FILE_TIME_LIMIT = 7 * 24 * 60 * 60

# CSP headers should be set, however this value can be used to disable them if necessary
ENABLE_CONTENT_SECURITY_POLICY = True

ENABLE_ELN_FILE_IMPORT = False
ELN_FILE_IMPORT_ALLOW_HTTP = False

ENABLE_FIDO2_PASSKEY_AUTHENTICATION = False

SHARED_DEVICE_SIGN_OUT_MINUTES = 30

DISABLE_OUTDATED_USE_AS_TEMPLATE = False

DISABLE_TOPICS = False

MIN_NUM_TEXT_CHOICES_FOR_SEARCH = 10

DEFAULT_NOTIFICATION_MODES = None

# environment variables override these values
use_environment_configuration(env_prefix='SAMPLEDB_')

if SERVICE_IMPRINT and not SERVICE_LEGAL_NOTICE:
    # Support SERVICE_IMPRINT for downwards compatibility
    SERVICE_LEGAL_NOTICE = SERVICE_IMPRINT

# remove trailing slashes from SciCat urls
if isinstance(SCICAT_API_URL, str) and SCICAT_API_URL.endswith('/'):
    SCICAT_API_URL = SCICAT_API_URL[:-1]  # pylint: disable=unsubscriptable-object
if isinstance(SCICAT_FRONTEND_URL, str) and SCICAT_FRONTEND_URL.endswith('/'):
    SCICAT_FRONTEND_URL = SCICAT_FRONTEND_URL[:-1]  # pylint: disable=unsubscriptable-object

# remove trailing slashes from Download Service url
if isinstance(DOWNLOAD_SERVICE_URL, str) and DOWNLOAD_SERVICE_URL.endswith('/'):
    DOWNLOAD_SERVICE_URL = DOWNLOAD_SERVICE_URL[:-1]  # pylint: disable=unsubscriptable-object

if OIDC_ONLY:
    DISABLE_INVITATIONS = True



# coding: utf-8
"""
Generate entries for the instrument_log_file_attachment_image_infos table.
"""

import flask_sqlalchemy

from ...logic.instrument_log_entries import generate_instrument_log_file_attachment_image_info


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    image_infos_exist = db.session.execute(db.text("""
        SELECT thumbnail_mime_type
        FROM instrument_log_file_attachment_image_infos
    """)).first() is not None
    if image_infos_exist:
        return False
    # Perform migration
    for file_attachment_id, in db.session.execute(db.text("""
        SELECT id
        FROM instrument_log_file_attachments
    """)).all():
        generate_instrument_log_file_attachment_image_info(file_attachment_id=file_attachment_id)
    return True



# coding: utf-8
"""
Logic module for management of instrument translations

Instrument translations complement instruments.
Translations contain all linguistic elements of instruments such as names.
"""

import dataclasses
import typing
from flask_babel import _

from .. import db
from .languages import Language
from .. import models
from . import errors, languages, instruments


@dataclasses.dataclass(frozen=True)
class InstrumentTranslation:
    """
    This class provides an immutable wrapper around models.instrument_translations.InstrumentTranslation.
    """
    instrument_id: int
    language_id: int
    name: str
    description: str
    short_description: str
    notes: str
    _language_cache: typing.List[languages.Language] = dataclasses.field(default_factory=list, kw_only=True, repr=False, compare=False)

    @classmethod
    def from_database(cls, instrument_translation: models.InstrumentTranslation) -> 'InstrumentTranslation':
        return InstrumentTranslation(
            instrument_id=instrument_translation.instrument_id,
            language_id=instrument_translation.language_id,
            name=instrument_translation.name or '',
            description=instrument_translation.description or '',
            short_description=instrument_translation.short_description or '',
            notes=instrument_translation.notes or ''
        )

    @property
    def language(self) -> Language:
        if not self._language_cache:
            self._language_cache.append(languages.get_language(self.language_id))
        return self._language_cache[0]


def set_instrument_translation(
        language_id: int,
        instrument_id: int,
        name: str,
        description: str,
        notes: str = '',
        short_description: str = ''
) -> InstrumentTranslation:
    """
    Creates a new instrument with the given name and description.

    :param language_id: the ID of an existing language
    :param instrument_id: the ID of an existing instrument
    :param name: the name of the instrument
    :param description: a (possibly empty) description of the instrument
    :param short_description: a (possibly empty) short description of the instrument
    :param notes: the notes shown to instrument responsible users
    :return: the new instrument translation
    """
    instrument_translation = models.InstrumentTranslation.query.filter_by(
        language_id=language_id,
        instrument_id=instrument_id
    ).first()
    if instrument_translation is None:
        instruments.check_instrument_exists(instrument_id)
        languages.get_language(language_id)
        instrument_translation = models.InstrumentTranslation(
            language_id=language_id,
            instrument_id=instrument_id,
            name=name,
            description=description,
            notes=notes,
            short_description=short_description,
        )
    else:
        instrument_translation.language_id = language_id
        instrument_translation.name = name
        instrument_translation.description = description
        instrument_translation.short_description = short_description
        instrument_translation.notes = notes
    db.session.add(instrument_translation)
    db.session.commit()
    return InstrumentTranslation.from_database(instrument_translation)


def get_instrument_translations_for_instrument(instrument_id: int, use_fallback: bool = False) -> typing.List[InstrumentTranslation]:
    """
    Returns all instrument translations for an instrument

    :param instrument_id: the ID of an existing instrument
    :param use_fallback: whether a fallback translation may be returned
    :return: a list of all translations for an instrument
    :raise errors.InstrumentTranslationDoesNotExistError: when there is no translation for the given instrument
    """
    translations = models.InstrumentTranslation.query.filter_by(instrument_id=instrument_id).order_by(models.InstrumentTranslation.language_id).all()
    if not translations and use_fallback:
        return [
            InstrumentTranslation(
                instrument_id=instrument_id,
                language_id=Language.ENGLISH,
                name=_('Unnamed Instrument (#%(instrument_id)s)', instrument_id=instrument_id),
                description='',
                short_description='',
                notes=''
            )
        ]
    return [
        InstrumentTranslation.from_database(instrument_translation)
        for instrument_translation in translations
    ]


def delete_instrument_translation(
        instrument_id: int,
        language_id: int
) -> None:
    """
    Delete an instrument translation.

    :param instrument_id: the ID of an existing instrument
    :param language_id: the ID of an existing language
    :raise errors.InstrumentTranslationDoesNotExistError: if there is no
        translation for the given instrument and language IDs
    """
    instrument_translation = models.InstrumentTranslation.query.filter_by(
        instrument_id=instrument_id,
        language_id=language_id
    ).first()
    if instrument_translation is None:
        raise errors.InstrumentTranslationDoesNotExistError()
    db.session.delete(instrument_translation)
    db.session.commit()



# coding: utf-8
"""
Logic module for locations and object location assignments.

Locations are fully user-defined using a name and a description. Users with
WRITE permissions for an object can assign it to a location with an optional
description for details on where the object is stored.
"""

import dataclasses
import datetime
import typing

from .components import Component
from . import actions, topics
from .action_types import check_action_type_exists
from .. import db, models, logic
from . import user_log, object_log, location_log, objects, users, errors, languages, components
from .notifications import create_notification_for_being_assigned_as_responsible_user
from ..models import locations


@dataclasses.dataclass(frozen=True)
class LocationType:
    """
    This class provides an immutable wrapper around models.locations.LocationType.
    """
    id: int
    name: typing.Optional[typing.Dict[str, str]]
    location_name_singular: typing.Optional[typing.Dict[str, str]]
    location_name_plural: typing.Optional[typing.Dict[str, str]]
    admin_only: bool
    enable_parent_location: bool
    enable_sub_locations: bool
    enable_object_assignments: bool
    enable_responsible_users: bool
    enable_instruments: bool
    enable_capacities: bool
    show_location_log: bool
    fed_id: typing.Optional[int] = None
    component_id: typing.Optional[int] = None
    component: typing.Optional[Component] = None

    LOCATION = locations.LocationType.LOCATION

    @classmethod
    def from_database(cls, location_type: locations.LocationType) -> 'LocationType':
        return LocationType(
            id=location_type.id,
            name=location_type.name,
            location_name_singular=location_type.location_name_singular,
            location_name_plural=location_type.location_name_plural,
            admin_only=location_type.admin_only,
            enable_parent_location=location_type.enable_parent_location,
            enable_sub_locations=location_type.enable_sub_locations,
            enable_object_assignments=location_type.enable_object_assignments,
            enable_responsible_users=location_type.enable_responsible_users,
            enable_instruments=location_type.enable_instruments,
            enable_capacities=location_type.enable_capacities,
            show_location_log=location_type.show_location_log,
            fed_id=location_type.fed_id,
            component_id=location_type.component_id,
            component=Component.from_database(location_type.component) if location_type.component is not None else None
        )


@dataclasses.dataclass(frozen=True)
class Location:
    """
    This class provides an immutable wrapper around models.locations.Location.
    """
    id: int
    name: typing.Optional[typing.Dict[str, str]]
    description: typing.Optional[typing.Dict[str, str]]
    type_id: int
    type: LocationType
    responsible_users: typing.List[users.User]
    is_hidden: bool
    enable_object_assignments: bool
    topics: typing.List[topics.Topic]
    parent_location_id: typing.Optional[int] = None
    fed_id: typing.Optional[int] = None
    component_id: typing.Optional[int] = None
    component: typing.Optional[Component] = None

    @classmethod
    def from_database(cls, location: locations.Location) -> 'Location':
        return Location(
            id=location.id,
            name=location.name,
            description=location.description,
            parent_location_id=location.parent_location_id,
            fed_id=location.fed_id,
            component=Component.from_database(location.component) if location.component is not None else None,
            component_id=location.component_id,
            type_id=location.type_id,
            type=LocationType.from_database(location.type),
            responsible_users=[
                users.User.from_database(responsible_user)
                for responsible_user in location.responsible_users
            ],
            is_hidden=location.is_hidden,
            enable_object_assignments=location.enable_object_assignments,
            topics=[topics.Topic.from_database(topic) for topic in location.topics],
        )


@dataclasses.dataclass(frozen=True)
class ObjectLocationAssignment:
    """
    This class provides an immutable wrapper around models.locations.ObjectLocationAssignment.
    """
    id: int
    object_id: int
    location_id: typing.Optional[int]
    user_id: typing.Optional[int]
    description: typing.Optional[typing.Dict[str, str]]
    utc_datetime: typing.Optional[datetime.datetime]
    responsible_user_id: typing.Optional[int]
    confirmed: bool
    fed_id: typing.Optional[int] = None
    component_id: typing.Optional[int] = None
    declined: bool = False
    component: typing.Optional[Component] = None

    @classmethod
    def from_database(cls, object_location_assignment: locations.ObjectLocationAssignment) -> 'ObjectLocationAssignment':
        return ObjectLocationAssignment(
            id=object_location_assignment.id,
            object_id=object_location_assignment.object_id,
            location_id=object_location_assignment.location_id,
            responsible_user_id=object_location_assignment.responsible_user_id,
            user_id=object_location_assignment.user_id,
            description=object_location_assignment.description,
            utc_datetime=object_location_assignment.utc_datetime,
            confirmed=object_location_assignment.confirmed,
            fed_id=object_location_assignment.fed_id,
            component_id=object_location_assignment.component_id,
            declined=object_location_assignment.declined,
            component=Component.from_database(object_location_assignment.component) if object_location_assignment.component is not None else None
        )


@typing.overload
def create_location(
        name: typing.Dict[str, str],
        description: typing.Dict[str, str],
        parent_location_id: typing.Optional[int],
        user_id: int,
        type_id: int,
        *,
        fed_id: None = None,
        component_id: None = None
) -> Location:
    ...


@typing.overload
def create_location(
        name: typing.Optional[typing.Dict[str, str]],
        description: typing.Optional[typing.Dict[str, str]],
        parent_location_id: typing.Optional[int],
        user_id: typing.Optional[int],
        type_id: int,
        *,
        fed_id: int,
        component_id: int
) -> Location:
    ...


def create_location(
        name: typing.Optional[typing.Dict[str, str]],
        description: typing.Optional[typing.Dict[str, str]],
        parent_location_id: typing.Optional[int],
        user_id: typing.Optional[int],
        type_id: int,
        *,
        fed_id: typing.Optional[int] = None,
        component_id: typing.Optional[int] = None
) -> Location:
    """
    Create a new location.

    :param name: the names for the new location in a dict. Keys are the language codes and values are the names.
    :param description: the descriptions for the new location.
        Keys are the language code and values are the descriptions.
    :param parent_location_id: the optional parent location ID for the new location
    :param user_id: the ID of an existing user
    :param fed_id: the federation ID of the location
    :param component_id: origin component ID
    :param type_id: the ID of an existing location type
    :return: the created location
    :raise errors.LocationDoesNotExistError: when no location with the given
        parent location ID exists
    :raise errors.UserDoesNotExistError: when no user with the given user ID
        exists
    :raise errors.LocationTypeDoesNotExistError: when no location type with the
        given location type ID exists
    :raise errors.MissingEnglishTranslationError: when name does not contain
        an english translation
    """
    assert (component_id is None) == (fed_id is None)
    assert component_id is not None or (name is not None and description is not None and user_id is not None)

    if name is not None:
        if isinstance(name, str):
            name = {
                'en': name
            }
        assert isinstance(name, dict)
        if component_id is None:
            name = languages.filter_translations(name)

        # if a name is provided, there should be an english translation
        if 'en' not in name:
            raise errors.MissingEnglishTranslationError()

    if description is not None:
        if isinstance(description, str):
            description = {
                'en': description
            }
        assert isinstance(description, dict)
        if component_id is None:
            description = languages.filter_translations(description)

    # ensure the user exists
    if user_id is not None:
        users.check_user_exists(user_id)
    if parent_location_id is not None:
        # ensure parent location exists
        check_location_exists(parent_location_id)
    if component_id is not None:
        # ensure that the component can be found
        components.get_component(component_id)
    if type_id is not None:
        # ensure location type exists
        get_location_type(type_id)
    location = locations.Location(
        name=name,
        description=description,
        parent_location_id=parent_location_id,
        fed_id=fed_id,
        component_id=component_id,
        type_id=type_id
    )
    db.session.add(location)
    db.session.commit()
    if component_id is None:
        if user_id is not None:
            user_log.create_location(user_id, location.id)
    location_log.create_location(user_id, location.id)
    return Location.from_database(location)


def update_location(
        location_id: int,
        name: typing.Optional[typing.Dict[str, str]],
        description: typing.Optional[typing.Dict[str, str]],
        parent_location_id: typing.Optional[int],
        user_id: typing.Optional[int],
        type_id: int,
        is_hidden: bool,
        enable_object_assignments: bool,
) -> None:
    """
    Update a location's information.

    :param location_id: the ID of the location to update
    :param name: the new names for the location in a dict. Keys are the language codes and the values the new names
    :param description: the descriptions for the location.
        Keys are the language code and the values the new descriptions.
    :param parent_location_id: the optional parent location id for the location
    :param user_id: the ID of an existing user
    :param type_id: the ID of an existing location type
    :param is_hidden: whether the location is hidden
    :param enable_object_assignments: whether object assignments are enabled for the location
    :raise errors.LocationDoesNotExistError: when no location with the given
        location ID or parent location ID exists
    :raise errors.CyclicLocationError: when location ID is an ancestor of
        parent location ID
    :raise errors.UserDoesNotExistError: when no user with the given user ID
        exists
    :raise errors.MissingEnglishTranslationError: if no english name is given
    :raise errors.LanguageDoesNotExistError: if an unknown language code is
        used
    :raise errors.LocationTypeDoesNotExistError: when no location type with the
        given location type ID exists
    """
    location = locations.Location.query.filter_by(id=location_id).first()
    if location is None:
        raise errors.LocationDoesNotExistError()

    if name is not None:
        assert isinstance(name, dict)
        if location.component_id is None:
            name = languages.filter_translations(name)

        # if a name is provided, there should be an english translation
        if 'en' not in name:
            raise errors.MissingEnglishTranslationError()

    if description is not None:
        assert isinstance(description, dict)
        if location.component_id is None:
            description = languages.filter_translations(description)

    # ensure the user exists
    if user_id is not None:
        users.check_user_exists(user_id)
    if parent_location_id is not None:
        if location_id == parent_location_id or location_id in _get_location_ancestors(parent_location_id):
            raise errors.CyclicLocationError()
    if type_id is not None:
        # ensure location type exists
        get_location_type(type_id)
    location.name = name
    location.description = description
    location.parent_location_id = parent_location_id
    location.type_id = type_id
    location.is_hidden = is_hidden
    location.enable_object_assignments = enable_object_assignments
    db.session.add(location)
    db.session.commit()
    if user_id is not None:
        user_log.update_location(user_id, location.id)
    location_log.update_location(user_id, location.id)


def check_location_exists(
        location_id: int
) -> None:
    """
    Check whether a location with the given location ID exists.

    :param location_id: the ID of an existing location
    :raise errors.LocationDoesNotExistError: when no location with the given
        location ID exists
    """
    if not db.session.query(db.exists().where(locations.Location.id == location_id)).scalar():
        raise errors.LocationDoesNotExistError()


def get_location(location_id: int, component_id: typing.Optional[int] = None) -> Location:
    """
    Get a location.

    :param location_id: the ID of an existing location
    :param component_id: the ID of an existing component, or None
    :raise errors.LocationDoesNotExistError: when no location with the given
        ID exists
    :return: the location with the given location ID
    """
    if component_id is None:
        location = locations.Location.query.filter_by(id=location_id).first()
    else:
        location = locations.Location.query.filter_by(fed_id=location_id, component_id=component_id).first()
    if location is None:
        if component_id is not None:
            components.check_component_exists(component_id)
        raise errors.LocationDoesNotExistError()
    return Location.from_database(location)


def get_locations() -> typing.List[Location]:
    """
    Get the list of all locations.

    :return: the list of all locations
    """
    return [
        Location.from_database(location)
        for location in locations.Location.query.all()
    ]


def get_locations_tree() -> typing.Tuple[typing.Dict[int, Location], typing.Any]:
    """
    Get the list of locations as a tree.

    :return: the list of all locations and the locations tree
    """
    locations = get_locations()
    locations_map = {
        location.id: location
        for location in locations
    }
    locations_tree: typing.Dict[typing.Optional[int], typing.Any] = {}
    locations_tree_helper: typing.Dict[typing.Optional[int], typing.Any] = {None: locations_tree}
    unvisited_locations = locations
    while unvisited_locations:
        locations = unvisited_locations
        unvisited_locations = []
        for location in locations:
            if location.parent_location_id in locations_tree_helper:
                locations_subtree = locations_tree_helper[location.parent_location_id]
                locations_subtree[location.id] = {}
                locations_tree_helper[location.id] = locations_subtree[location.id]
            else:
                unvisited_locations.append(location)
    return locations_map, locations_tree


def is_full_location_tree_hidden(
    locations_map: typing.Dict[int, Location],
    locations_tree: typing.Any
) -> bool:
    """
    Return whether a full locations tree only contains hidden locations.

    :param locations_map: a dict mapping location IDs to locations
    :param locations_tree: the locations tree
    :return: whether all locations in the tree are hidden
    """
    return all(
        locations_map[location_id].is_hidden and is_full_location_tree_hidden(locations_map, locations_subtree)
        for location_id, locations_subtree in locations_tree.items()
        if location_id in locations_map
    )


def _get_location_ancestors(location_id: int) -> typing.List[int]:
    """
    Get the list of all ancestor location IDs.

    :param location_id: the ID of an existing location
    :return: the list of all ancestor location IDs
    :raise errors.LocationDoesNotExistError: when no location with the given
        location ID exists
    """
    ancestor_location_ids = []
    ancestor_location_id: typing.Optional[int] = location_id
    while ancestor_location_id is not None:
        ancestor_location_ids.append(ancestor_location_id)
        ancestor_location_id = get_location(ancestor_location_id).parent_location_id
    return ancestor_location_ids[1:]


def assign_location_to_object(
        object_id: int,
        location_id: typing.Optional[int],
        responsible_user_id: typing.Optional[int],
        user_id: int,
        description: typing.Optional[typing.Dict[str, str]]
) -> None:
    """
    Assign a location to an object.

    :param object_id: the ID of an existing object
    :param location_id: the ID of an existing location
    :param responsible_user_id: the ID of an existing user
    :param user_id: the ID of an existing user
    :param description: a description of where the object was stored in a dict.
        The keys are the lang codes and the values are the descriptions.
    :raise errors.ObjectDoesNotExistError: when no object with the given
        object ID exists
    :raise errors.LocationDoesNotExistError: when no location with the given
        location ID exists
    :raise errors.UserDoesNotExistError: when no user with the given user ID
        or responsible user ID exists
    :raise errors.ExceedingLocationCapacityError: when assigning the object
        to this location would exceed the location's capacity
    """

    if description is not None:
        if isinstance(description, str):
            description = {
                'en': description
            }
        assert isinstance(description, dict)
        description = languages.filter_translations(description)

    # ensure the object exists
    objects.check_object_exists(object_id)
    if location_id is not None:
        location = get_location(location_id)
        if location.type.enable_capacities:
            action_id = objects.get_object(object_id).action_id
            if action_id is None:
                raise errors.ExceedingLocationCapacityError()
            action_type_id = actions.get_action(action_id).type_id
            if action_type_id is None:
                raise errors.ExceedingLocationCapacityError()
            capacity = get_location_capacities(location_id).get(action_type_id, 0)
            if capacity is not None:
                if capacity == 0:
                    raise errors.ExceedingLocationCapacityError()
                num_stored_objects = get_assigned_object_count_by_action_types(
                    location_id,
                    ignored_object_ids=[object_id]
                ).get(action_type_id, 0)
                if num_stored_objects + 1 > capacity:
                    raise errors.ExceedingLocationCapacityError()
    # ensure the user exists
    users.check_user_exists(user_id)
    object_location_assignment = locations.ObjectLocationAssignment(
        object_id=object_id,
        location_id=location_id,
        responsible_user_id=responsible_user_id,
        user_id=user_id,
        description=description,
        utc_datetime=datetime.datetime.now(datetime.timezone.utc),
        confirmed=(user_id == responsible_user_id),
        declined=False
    )
    db.session.add(object_location_assignment)
    db.session.commit()
    if responsible_user_id is not None:
        users.check_user_exists(responsible_user_id)
        if user_id != responsible_user_id:
            create_notification_for_being_assigned_as_responsible_user(object_location_assignment.id)
    object_log.assign_location(user_id, object_id, object_location_assignment.id)
    user_log.assign_location(user_id, object_location_assignment.id)
    previous_object_location_assignment = locations.ObjectLocationAssignment.query.filter_by(
        object_id=object_id
    ).order_by(
        db.desc(locations.ObjectLocationAssignment.utc_datetime)
    ).offset(1).first()
    if previous_object_location_assignment is not None:
        previous_location_id = previous_object_location_assignment.location_id
    else:
        previous_location_id = None
    if location_id is not None and location_id == previous_location_id:
        location_log.change_object(user_id, location_id, object_location_assignment.id)
    else:
        if location_id is not None:
            location_log.add_object(user_id, location_id, object_location_assignment.id)
        if previous_location_id is not None:
            location_log.remove_object(user_id, previous_location_id, object_location_assignment.id)


def create_fed_assignment(
        fed_id: int,
        component_id: int,
        object_id: int,
        location_id: typing.Optional[int],
        responsible_user_id: typing.Optional[int],
        user_id: typing.Optional[int],
        description: typing.Optional[typing.Dict[str, str]],
        utc_datetime: typing.Optional[datetime.datetime],
        confirmed: bool = False,
        declined: bool = False
) -> locations.ObjectLocationAssignment:
    if description is not None:
        if isinstance(description, str):
            description = {
                'en': description
            }
        assert isinstance(description, dict)
        description = languages.filter_translations(description)

    objects.check_object_exists(object_id)
    # ensure the component exists
    components.check_component_exists(component_id)
    if location_id is not None:
        # ensure the location exists
        check_location_exists(location_id)
    if user_id is not None:
        users.check_user_exists(user_id)
    if responsible_user_id is not None:
        users.check_user_exists(responsible_user_id)
    object_location_assignment = locations.ObjectLocationAssignment(
        object_id=object_id,
        location_id=location_id,
        responsible_user_id=responsible_user_id,
        user_id=user_id,
        description=description,
        utc_datetime=utc_datetime,
        confirmed=confirmed or False,
        declined=declined,
        fed_id=fed_id,
        component_id=component_id
    )
    db.session.add(object_location_assignment)
    db.session.commit()
    return object_location_assignment


def get_object_location_assignments(object_id: int) -> typing.List[ObjectLocationAssignment]:
    """
    Get a list of all object location assignments for an object.

    :param object_id: the ID of an existing object
    :return: the list of object location assignments
    :raise errors.ObjectDoesNotExistError: when no object with the given
        object ID exists
    """
    # ensure the object exists
    objects.check_object_exists(object_id)
    object_location_assignments = locations.ObjectLocationAssignment.query.filter_by(object_id=object_id).order_by(locations.ObjectLocationAssignment.utc_datetime).all()
    return [
        ObjectLocationAssignment.from_database(object_location_assignment)
        for object_location_assignment in object_location_assignments
    ]


def get_fed_object_location_assignment(fed_id: int, component_id: int) -> typing.Optional[locations.ObjectLocationAssignment]:
    object_location_assignment: typing.Optional[locations.ObjectLocationAssignment] = locations.ObjectLocationAssignment.query.filter_by(fed_id=fed_id, component_id=component_id).first()
    return object_location_assignment


def get_current_object_location_assignment(object_id: int) -> typing.Optional[ObjectLocationAssignment]:
    """
    Get the current location assignment for an object.

    :param object_id: the ID of an existing object
    :return: the object location assignment
    :raise errors.ObjectDoesNotExistError: when no object with the given
        object ID exists
    """
    # ensure the object exists
    objects.check_object_exists(object_id)
    object_location_assignment: typing.Optional[locations.ObjectLocationAssignment] = locations.ObjectLocationAssignment.query.filter_by(object_id=object_id).order_by(locations.ObjectLocationAssignment.utc_datetime.desc()).first()
    if object_location_assignment is None:
        return None
    return ObjectLocationAssignment.from_database(object_location_assignment)


def get_current_object_location_assignments(
        object_ids: typing.Sequence[int]
) -> typing.Mapping[int, typing.Optional[ObjectLocationAssignment]]:
    """
    Get the current location assignments for a list of objects.

    :param object_ids: a list of IDs of existing objects
    :return: a dict mapping object IDs to the object location assignments
    :raise errors.ObjectDoesNotExistError: when no object with one of the
        given object IDs exists
    """

    assignment_rows = locations.ObjectLocationAssignment.query.distinct(
        locations.ObjectLocationAssignment.object_id
    ).filter(
        locations.ObjectLocationAssignment.object_id.in_(object_ids)
    ).order_by(
        locations.ObjectLocationAssignment.object_id, locations.ObjectLocationAssignment.utc_datetime.desc()
    ).all()

    object_location_assignments: typing.Dict[int, typing.Optional[ObjectLocationAssignment]] = {}
    for assignment_row in assignment_rows:
        object_location_assignments[assignment_row.object_id] = ObjectLocationAssignment.from_database(assignment_row)
    for object_id in object_ids:
        if object_id not in object_location_assignments:
            objects.check_object_exists(object_id)
            object_location_assignments[object_id] = None
    return object_location_assignments


def _get_mutable_object_location_assignment(object_location_assignment_id: int) -> locations.ObjectLocationAssignment:
    """
    Get a mutable object location assignment with a given ID.

    :param object_location_assignment_id: the ID of an existing object location
        assignment
    :return: the mutable object location assignment
    :raise errors.ObjectLocationAssignmentDoesNotExistError: when no object
        location assignment with the given object location assignment ID exists
    """
    object_location_assignment = locations.ObjectLocationAssignment.query.filter_by(id=object_location_assignment_id).first()
    if object_location_assignment is None:
        raise errors.ObjectLocationAssignmentDoesNotExistError()
    return object_location_assignment


def get_object_location_assignment(object_location_assignment_id: int) -> ObjectLocationAssignment:
    """
    Get an object location assignment with a given ID.

    :param object_location_assignment_id: the ID of an existing object location
        assignment
    :return: the object location assignment
    :raise errors.ObjectLocationAssignmentDoesNotExistError: when no object
        location assignment with the given object location assignment ID exists
    """
    return ObjectLocationAssignment.from_database(_get_mutable_object_location_assignment(object_location_assignment_id))


def any_objects_at_location(location_id: int) -> bool:
    """
    Get whether there are any objects currently assigned to a location.

    :param location_id: the ID of an existing location
    :return: whether there are any objects assigned to the location
    :raise errors.LocationDoesNotExistError: when no location with the given
        location ID exists
    """
    # ensure the location exists
    check_location_exists(location_id)
    object_location_assignments = locations.ObjectLocationAssignment.query.filter_by(location_id=location_id).all()
    for object_location_assignment in object_location_assignments:
        object_id = object_location_assignment.object_id
        current_object_location_assignment = get_current_object_location_assignment(object_id)
        if current_object_location_assignment is not None and current_object_location_assignment.location_id == location_id:
            return True
    return False


def get_object_ids_at_location(location_id: int) -> typing.Set[int]:
    """
    Get a list of all objects currently assigned to a location.

    :param location_id: the ID of an existing location
    :return: the list of object IDs assigned to the location
    :raise errors.LocationDoesNotExistError: when no location with the given
        location ID exists
    """
    # ensure the location exists
    check_location_exists(location_id)
    object_location_assignments = locations.ObjectLocationAssignment.query.filter_by(location_id=location_id).all()
    object_ids = set()
    for object_location_assignment in object_location_assignments:
        object_id = object_location_assignment.object_id
        current_object_location_assignment = get_current_object_location_assignment(object_id)
        if current_object_location_assignment is not None and current_object_location_assignment.location_id == location_id:
            object_ids.add(object_id)
    return object_ids


def confirm_object_responsibility(object_location_assignment_id: int) -> None:
    """
    Confirm an object location assignment.

    :param object_location_assignment_id: the ID of an existing object location
        assignment
    :raise errors.ObjectLocationAssignmentDoesNotExistError: when no object
        location assignment with the given object location assignment ID exists
    :raise errors.ObjectLocationAssignmentAlreadyDeclinedError: when the
        object location assignment has already been declined
    """
    object_location_assignment = _get_mutable_object_location_assignment(object_location_assignment_id)
    if object_location_assignment.declined:
        raise errors.ObjectLocationAssignmentAlreadyDeclinedError()
    object_location_assignment.confirmed = True
    db.session.add(object_location_assignment)
    db.session.commit()


def decline_object_responsibility(object_location_assignment_id: int) -> None:
    """
    Decline responsibility for an object.

    :param object_location_assignment_id: the ID of an existing object location
        assignment
    :raise errors.ObjectLocationAssignmentDoesNotExistError: when no object
        location assignment with the given object location assignment ID exists
    :raise errors.ObjectLocationAssignmentAlreadyConfirmedError: when the
        object location assignment has already been confirmed
    """
    object_location_assignment = _get_mutable_object_location_assignment(object_location_assignment_id)
    if object_location_assignment.confirmed:
        raise errors.ObjectLocationAssignmentAlreadyConfirmedError()
    object_location_assignment.declined = True
    db.session.add(object_location_assignment)
    db.session.commit()


def get_unhandled_object_responsibility_assignments(
        user_id: int
) -> typing.List[ObjectLocationAssignment]:
    """
    Get all object location assignments that assign responsibility to a given
    user and have not yet been confirmed or declined.

    :param user_id: the ID of an existing user
    :return: the object location assignments
    """
    object_location_assignments = locations.ObjectLocationAssignment.query.filter_by(
        responsible_user_id=user_id,
        confirmed=False,
        declined=False
    ).order_by(locations.ObjectLocationAssignment.utc_datetime.desc()).all()
    return [
        ObjectLocationAssignment.from_database(object_location_assignment)
        for object_location_assignment in object_location_assignments
    ]


def create_location_type(
        name: typing.Optional[typing.Dict[str, str]],
        location_name_singular: typing.Optional[typing.Dict[str, str]],
        location_name_plural: typing.Optional[typing.Dict[str, str]],
        admin_only: bool,
        enable_parent_location: bool,
        enable_sub_locations: bool,
        enable_object_assignments: bool,
        enable_responsible_users: bool,
        enable_instruments: bool,
        enable_capacities: bool,
        show_location_log: bool,
        fed_id: typing.Optional[int] = None,
        component_id: typing.Optional[int] = None
) -> LocationType:
    """
    Create a new location type.

    :param name: the names for the new location type in a dict. Keys are the
        language codes and values are the names.
    :param location_name_singular: the names for single locations of this type in a dict
    :param location_name_plural: the names for multiple locations of this type in a dict
    :param admin_only: whether only administrators can create locations of this type
    :param enable_parent_location: whether locations of this type may have a parent location
    :param enable_sub_locations: whether locations of this type may have sub locations
    :param enable_object_assignments: whether objects may be assigned to locations of this type
    :param enable_responsible_users: whether locations of this type may have responsible users
    :param enable_instruments: whether instruments may be assigned to locations of this type
    :param enable_capacities: whether locations of this type have capacities
    :param show_location_log: whether the location log should be shown for locations of this type
    :param fed_id: the federation ID of the location type
    :param component_id: origin component ID
    :return: the created location type
    """
    location_type = locations.LocationType(
        name=name,
        location_name_singular=location_name_singular,
        location_name_plural=location_name_plural,
        admin_only=admin_only,
        enable_parent_location=enable_parent_location,
        enable_sub_locations=enable_sub_locations,
        enable_object_assignments=enable_object_assignments,
        enable_responsible_users=enable_responsible_users,
        enable_instruments=enable_instruments,
        enable_capacities=enable_capacities,
        show_location_log=show_location_log,
        fed_id=fed_id,
        component_id=component_id
    )
    db.session.add(location_type)
    db.session.commit()
    return LocationType.from_database(location_type)


def update_location_type(
        location_type_id: int,
        name: typing.Optional[typing.Dict[str, str]],
        location_name_singular: typing.Optional[typing.Dict[str, str]],
        location_name_plural: typing.Optional[typing.Dict[str, str]],
        admin_only: bool,
        enable_parent_location: bool,
        enable_sub_locations: bool,
        enable_object_assignments: bool,
        enable_responsible_users: bool,
        enable_instruments: bool,
        enable_capacities: bool,
        show_location_log: bool,
) -> None:
    """
    Create a new location type.

    :param location_type_id: the ID of an existing location type
    :param name: the names for the new location type in a dict. Keys are the
        language codes and values are the names.
    :param location_name_singular: the names for single locations of this type in a dict
    :param location_name_plural: the names for multiple locations of this type in a dict
    :param admin_only: whether only administrators can create locations of this type
    :param enable_parent_location: whether locations of this type may have a parent location
    :param enable_sub_locations: whether locations of this type may have sub locations
    :param enable_object_assignments: whether objects may be assigned to locations of this type
    :param enable_responsible_users: whether locations of this type may have responsible users
    :param enable_instruments: whether instruments may be assigned to locations of this type
    :param enable_capacities: whether locations of this type have capacities
    :param show_location_log: whether the location log should be shown for locations of this type
    :raise errors.LocationTypeDoesNotExistError: if no location type with the
        given ID exists
    """
    location_type = locations.LocationType.query.filter_by(id=location_type_id).first()
    if location_type is None:
        raise errors.LocationTypeDoesNotExistError()
    location_type.name = name
    location_type.location_name_singular = location_name_singular
    location_type.location_name_plural = location_name_plural
    location_type.admin_only = admin_only
    location_type.enable_parent_location = enable_parent_location
    location_type.enable_sub_locations = enable_sub_locations
    location_type.enable_object_assignments = enable_object_assignments
    location_type.enable_responsible_users = enable_responsible_users
    location_type.enable_instruments = enable_instruments
    location_type.enable_capacities = enable_capacities
    location_type.show_location_log = show_location_log
    db.session.add(location_type)
    db.session.commit()


def get_location_type(
        location_type_id: int,
        component_id: typing.Optional[int] = None
) -> LocationType:
    """
    Get a location type.

    :param location_type_id: the ID of an existing location type
    :param component_id: the ID of an existing component, or None
    :raise errors.LocationTypeDoesNotExistError: when no location type with
        the given ID exists
    :return: the location type with the given location type ID
    """
    if component_id is None:
        location_type = locations.LocationType.query.filter_by(id=location_type_id).first()
    else:
        location_type = locations.LocationType.query.filter_by(fed_id=location_type_id, component_id=component_id).first()
    if location_type is None:
        if component_id is not None:
            components.check_component_exists(component_id)
        raise errors.LocationTypeDoesNotExistError()
    return LocationType.from_database(location_type)


def get_location_types() -> typing.List[LocationType]:
    """
    Get all location types.

    :return: the list of all location types
    """
    location_types = locations.LocationType.query.order_by(db.asc(locations.LocationType.id)).all()
    return [
        LocationType.from_database(location_type)
        for location_type in location_types
    ]


def set_location_responsible_users(
        location_id: int,
        responsible_user_ids: typing.List[int]
) -> None:
    """
    Set the responsible users for a location.

    :param location_id: the ID of an existing location
    :param responsible_user_ids: a list of user IDs
    :raise errors.LocationDoesNotExistError: when no location with the given
        location ID exists
    :raise errors.UserDoesNotExistError: when no user with the given
        user ID exists
    """
    location = locations.Location.query.filter_by(id=location_id).first()
    if location is None:
        raise errors.LocationDoesNotExistError()
    location.responsible_users.clear()
    for user_id in responsible_user_ids:
        user = users.get_mutable_user(user_id)
        location.responsible_users.append(user)
    db.session.add(location)
    db.session.commit()


def get_descendent_location_ids(
        locations_tree: typing.Dict[typing.Optional[int], typing.Any]
) -> typing.Set[int]:
    location_tree_queue = [locations_tree]
    descendent_location_ids: typing.Set[int] = set()
    while location_tree_queue:
        locations_tree = location_tree_queue.pop()
        for child_location_id in locations_tree:
            if child_location_id is not None:
                descendent_location_ids.add(child_location_id)
                location_tree_queue.append(locations_tree[child_location_id])
    return descendent_location_ids


def get_location_capacities(
        location_id: int
) -> typing.Dict[int, typing.Optional[int]]:
    """
    Get the capacity by action type for a given location.

    :param location_id: the ID of an existing location
    :return: a dict mapping action type IDs to capacities
    :raise errors.LocationDoesNotExistError: when no location with the given
        location ID exists
    """
    location_capacities = locations.LocationCapacity.query.filter_by(location_id=location_id).all()
    if not location_capacities:
        check_location_exists(location_id)
    return {
        location_capacity.action_type_id: location_capacity.capacity
        for location_capacity in location_capacities
    }


def set_location_capacity(
        location_id: int,
        action_type_id: int,
        capacity: typing.Optional[int]
) -> None:
    """
    Set the capacity for objects of a given action type at a given location.

    :param location_id: the ID of an existing location
    :param action_type_id: the ID of an existing action type
    :param capacity: the storage capacity, or None for unlimited capacity
    :raise errors.LocationDoesNotExistError: when no location with the given
        location ID exists
    :raise errors.ActionTypeDoesNotExistError: when no action type with the
        given action type ID exists
    """
    check_location_exists(location_id)
    check_action_type_exists(action_type_id)
    if capacity is not None and capacity < 0:
        capacity = 0
    location_capacity = locations.LocationCapacity.query.filter_by(location_id=location_id, action_type_id=action_type_id).first()
    if capacity == 0:
        if location_capacity is not None:
            db.session.delete(location_capacity)
    else:
        if location_capacity is None:
            location_capacity = locations.LocationCapacity(
                location_id=location_id,
                action_type_id=action_type_id,
                capacity=capacity
            )
        else:
            location_capacity.capacity = capacity
        db.session.add(location_capacity)
    db.session.commit()


def clear_location_capacities(
        location_id: int
) -> None:
    """
    Set the capacity for objects of any action type at a given location to 0.

    :param location_id: the ID of an existing location
    :raise errors.LocationDoesNotExistError: when no location with the given
        location ID exists
    """
    check_location_exists(location_id)
    location_capacities = locations.LocationCapacity.query.filter_by(location_id=location_id).all()
    if location_capacities:
        for location_capacity in location_capacities:
            db.session.delete(location_capacity)
        db.session.commit()
    else:
        check_location_exists(location_id)


def get_assigned_object_count_by_action_types(  # pylint: disable=useless-param-doc
        location_id: int,
        ignored_object_ids: typing.Sequence[int] = ()
) -> typing.Dict[int, int]:
    """
    Get the amount of objects (by action type) assigned to a given location.

    :param location_id: the ID of an existing location
    :param ignored_object_ids: a list of object IDs to ignore
    :return: a dict mapping action type IDs to object counts
    :raise errors.LocationDoesNotExistError: when no location with the given
        location ID exists
    """
    # location -> objects
    location_assignments = locations.ObjectLocationAssignment.query.filter_by(
        location_id=location_id
    ).filter(
        db.not_(locations.ObjectLocationAssignment.object_id.in_(ignored_object_ids))
    ).subquery('location_assignments')
    current_location_assignments = locations.ObjectLocationAssignment.query.distinct(
        locations.ObjectLocationAssignment.object_id
    ).filter(
        locations.ObjectLocationAssignment.object_id == location_assignments.c.object_id
    ).order_by(
        locations.ObjectLocationAssignment.object_id,
        locations.ObjectLocationAssignment.utc_datetime.desc()
    ).all()
    current_location_assignments = [
        location_assignment
        for location_assignment in current_location_assignments
        if location_assignment.location_id == location_id
    ]
    if not current_location_assignments:
        check_location_exists(location_id)
        return {}
    # objects -> actions
    stored_object_ids = [
        a.object_id for a in current_location_assignments
    ]
    action_ids_for_object_ids = logic.objects.get_action_ids_for_object_ids(stored_object_ids)
    # actions -> action types
    stored_action_ids = [
        typing.cast(int, action_ids_for_object_ids[object_id])
        for object_id in stored_object_ids
        if action_ids_for_object_ids.get(object_id) is not None
    ]
    action_type_ids_for_action_ids = logic.actions.get_action_type_ids_for_action_ids(stored_action_ids)
    # action types -> fed action types
    stored_action_type_ids = set(action_type_ids_for_action_ids.values())
    action_type_ids_and_fed_action_type_ids = models.actions.ActionType.query.with_entities(
        models.actions.ActionType.id,
        models.actions.ActionType.fed_id
    ).filter(
        models.actions.ActionType.fed_id.is_not(None),
        models.actions.ActionType.fed_id < 0,
        models.actions.ActionType.id.in_(stored_action_type_ids)
    ).all()
    fed_action_type_ids_for_action_type_ids = {
        action_type_id: fed_action_type_id
        for action_type_id, fed_action_type_id in action_type_ids_and_fed_action_type_ids
    }
    # location -> action types
    result = {}
    for object_id in stored_object_ids:
        action_id = action_ids_for_object_ids.get(object_id)
        if action_id is None:
            continue
        action_type_id = action_type_ids_for_action_ids.get(action_id)
        if action_type_id is None:
            continue
        action_type_id = fed_action_type_ids_for_action_type_ids.get(action_type_id, action_type_id)
        if action_type_id is None:
            continue
        if action_type_id not in result:
            result[action_type_id] = 0
        result[action_type_id] += 1
    return result



import base64
import datetime
import hashlib
import json
import mimetypes
import os.path
import typing
import urllib.parse

import flask
import minisign
from flask import url_for

from . import minisign_keys, markdown_images
from .files import File
from .utils import get_translated_text
from .actions import get_action
from .action_types import ActionType
from .objects import find_object_references
from .dataverse_export import flatten_metadata, get_title_for_property
from .datatypes import Quantity
from .units import get_un_cefact_code_for_unit
from .users import get_user
from .. import version


def _unpack_single_item_arrays(json_value: typing.Any) -> typing.Any:
    if isinstance(json_value, dict):
        return {
            key: _unpack_single_item_arrays(value)
            for key, value in json_value.items()
        }
    if isinstance(json_value, (list, tuple)):
        if len(json_value) == 1:
            return _unpack_single_item_arrays(json_value[0])
        return [
            _unpack_single_item_arrays(value)
            for value in json_value
        ]
    return json_value


def generate_ro_crate_metadata(
        archive_files: typing.Dict[str, typing.Union[str, bytes]],
        infos: typing.Dict[str, typing.Any],
        user_id: int,
        object_ids: typing.Optional[typing.List[int]]
) -> typing.Dict[str, bytes]:
    result_files: typing.Dict[str, bytes] = {}
    date_created = datetime.datetime.now(datetime.timezone.utc).replace(tzinfo=None).isoformat(timespec='microseconds')
    description = f"SampleDB .eln export generated for user #{user_id}"
    if object_ids:
        description += " for objects #" + ', #'.join(map(str, sorted(object_ids)))
    ro_crate_metadata: typing.Dict[str, typing.Any] = {
        "@context": "https://w3id.org/ro/crate/1.2/context",
        "@graph": [
            {
                "@type": "CreativeWork",
                "@id": "ro-crate-metadata.json",
                "description": "RO-Crate Metadata File Descriptor",
                "about": {
                    "@id": "./"
                },
                "conformsTo": {
                    "@id": "https://w3id.org/ro/crate/1.2"
                },
                "sdPublisher": {
                    "@id": "SampleDB"
                },
                "version": "1.2",
                "dateCreated": date_created
            },
            {
                "@id": "./",
                "@type": "Dataset",
                "name": "SampleDB .eln export",
                "description": description,
                "license": {
                    "@id": "./license"
                },
                "datePublished": date_created,
                "hasPart": []
            },
            {
                "@type": "Organization",
                "@id": "SampleDB",
                "name": "SampleDB",
                "logo": "https://scientific-it-systems.iffgit.fz-juelich.de/SampleDB/_static/img/logo.svg",
                "slogan": "SampleDB is a web-based electronic lab notebook with a focus on sample and measurement metadata.",
                "url": "https://scientific-it-systems.iffgit.fz-juelich.de/SampleDB/"
            },
            {
                "@id": "./license",
                "@type": "CreativeWork",
                "name": "No License",
                "description": "This .eln file does not include a license."
            }
        ]
    }
    directory_datasets = {
        "sampledb_export": ro_crate_metadata["@graph"][1]
    }
    exported_object_ids = {
        object_info['id']: f"./objects/{object_info['id']}/"
        for object_info in infos['objects']
    }
    exported_user_ids = {
        user_info['id']: f"./users/{user_info['id']}"
        for user_info in infos['users']
    }
    for object_info in infos['objects']:
        if object_info['action_id'] is not None:
            action = get_action(object_info['action_id'])
            if action.type_id is None:
                object_type = None
            else:
                object_type = {
                    ActionType.SAMPLE_CREATION: "sample",
                    ActionType.MEASUREMENT: "measurement",
                    ActionType.SIMULATION: "simulation",
                }.get(action.type_id, None)
        else:
            object_type = 'other'
        exported_file_ids = {
            file_info['id']: './' + file_info['path']
            for file_info in object_info['files']
            if 'path' in file_info
        }
        property_values = _convert_metadata_to_property_values(
            object_id=object_info['id'],
            data=object_info['versions'][0]['data'],
            schema=object_info['versions'][0]['schema'],
            exported_object_ids=exported_object_ids,
            exported_user_ids=exported_user_ids,
            exported_file_ids=exported_file_ids,
            property_id_prefix=f"./objects/{object_info['id']}/properties/"
        ) if object_info['versions'][0]['data'] is not None else []
        ro_crate_metadata["@graph"].extend(property_values)
        property_value_references = [
            {
                '@id': property_value['@id']
            }
            for property_value in property_values
        ]
        ro_crate_metadata["@graph"].append({
            "@id": f"./objects/{object_info['id']}/",
            "@type": "Dataset",
            "identifier": f"{object_info['id']}",
            "name": f"{get_translated_text(object_info['versions'][-1]['data'].get('name', {}).get('text', {}), 'en')}" if object_info['versions'][-1]['data'] is not None else '',
            "description": f"Object #{object_info['id']}",
            "dateCreated": object_info['versions'][0]['utc_datetime'],
            "dateModified": object_info['versions'][-1]['utc_datetime'],
            "author": [
                {"@id": f"./users/{version_user_id}"}
                for version_user_id in sorted({
                    version_info['user_id']
                    for version_info in object_info['versions']
                    if version_info['user_id'] is not None
                })
            ],
            "creator": {"@id": f"./users/{object_info['versions'][0]['user_id']}"} if object_info['versions'][0]['user_id'] is not None else None,
            "url": flask.url_for('frontend.object', object_id=object_info['id'], _external=True),
            "variableMeasured": property_value_references,
            "mentions": [],
            "comment": [],
            "isBasedOn": [
                {
                    "@id": f"./objects/{object_info['id']}/versions/{version_info['id']}/",
                }
                for version_info in object_info['versions']
            ],
            "hasPart": [
                {
                    "@id": f"./objects/{object_info['id']}/versions/{version_info['id']}/",
                }
                for version_info in object_info['versions']
            ] + [
                {
                    "@id": f"./objects/{object_info['id']}/files.json"
                }
            ]
        })
        if object_type:
            ro_crate_metadata["@graph"][-1]["genre"] = object_type
        current_version_info = object_info['versions'][-1]
        if current_version_info.get('data') and isinstance(current_version_info['data'].get('tags'), dict) and current_version_info['data']['tags'].get('_type') == 'tags' and current_version_info['data']['tags'].get('tags'):
            ro_crate_metadata["@graph"][-1]["keywords"] = ', '.join(current_version_info['data']['tags']['tags'])
        if current_version_info.get('data'):
            referenced_object_ids = find_object_references(
                object_id=typing.cast(int, object_info['id']),
                version_id=typing.cast(int, current_version_info['id']),
                object_data=typing.cast(typing.Optional[typing.Dict[str, typing.Any]], current_version_info['data']),
                find_previous_referenced_object_ids=False,
                include_fed_references=False
            )
            for referenced_object_id, _previously_reference_object_id, _object_reference_type in referenced_object_ids:
                if referenced_object_id in exported_object_ids:
                    ro_crate_metadata["@graph"][-1]["mentions"].append({"@id": exported_object_ids[referenced_object_id]})
        if not ro_crate_metadata["@graph"][-1]["mentions"]:
            del ro_crate_metadata["@graph"][-1]["mentions"]
        directory_datasets[f"sampledb_export/objects/{object_info['id']}/files"] = ro_crate_metadata["@graph"][-1]
        directory_datasets[f"sampledb_export/objects/{object_info['id']}/comments"] = ro_crate_metadata["@graph"][-1]
        directory_datasets["sampledb_export"]["hasPart"].append({
            "@id": f"./objects/{object_info['id']}/"
        })
        for version_info in object_info['versions']:
            property_values = _convert_metadata_to_property_values(
                object_id=object_info['id'],
                data=version_info['data'],
                schema=version_info['schema'],
                exported_object_ids=exported_object_ids,
                exported_user_ids=exported_user_ids,
                exported_file_ids=exported_file_ids,
                property_id_prefix=f"./objects/{object_info['id']}/versions/{version_info['id']}/properties/"
            ) if version_info['data'] is not None else []
            ro_crate_metadata["@graph"].extend(property_values)
            property_value_references = [
                {
                    '@id': property_value['@id']
                }
                for property_value in property_values
            ]
            ro_crate_metadata["@graph"].append({
                "@id": f"./objects/{object_info['id']}/versions/{version_info['id']}/",
                "@type": "Dataset",
                "name": f"{get_translated_text(version_info['data'].get('name', {}).get('text', {}), 'en')}" if version_info['data'] is not None else '',
                "description": f"Object #{object_info['id']} version #{version_info['id']}",
                "dateCreated": version_info['utc_datetime'],
                "creator": {"@id": f"./users/{version_info['user_id']}"} if version_info['user_id'] is not None else None,
                "author": [
                    {"@id": f"./users/{other_version_user_id}"}
                    for other_version_user_id in sorted({
                        other_version_info['user_id']
                        for other_version_info in object_info['versions']
                        if other_version_info['user_id'] is not None and other_version_info['id'] <= version_info['id']
                    })
                ],
                "url": flask.url_for('frontend.object_version', object_id=object_info['id'], version_id=version_info['id'], _external=True),
                "variableMeasured": property_value_references,
                "hasPart": [
                    {
                        "@id": f"./objects/{object_info['id']}/versions/{version_info['id']}/schema.json",
                    },
                    {
                        "@id": f"./objects/{object_info['id']}/versions/{version_info['id']}/data.json"
                    }
                ]
            })
            previous_version_refs = []
            for other_version_info in object_info['versions']:
                if other_version_info['id'] < version_info['id']:
                    previous_version_refs.append({
                        "@id": f"./objects/{object_info['id']}/versions/{other_version_info['id']}/"
                    })
            if previous_version_refs:
                ro_crate_metadata["@graph"][-1]['isBasedOn'] = previous_version_refs

            schema_json = json.dumps(version_info['schema'], indent=2).encode('utf-8')
            result_files[f"sampledb_export/objects/{object_info['id']}/versions/{version_info['id']}/schema.json"] = schema_json
            ro_crate_metadata["@graph"].append({
                "@id": f"./objects/{object_info['id']}/versions/{version_info['id']}/schema.json",
                "@type": "File",
                "description": f"Schema for Object #{object_info['id']} version #{version_info['id']}",
                "name": "schema.json",
                "encodingFormat": "application/json",
                "contentSize": str(len(schema_json)),
                "sha256": hashlib.sha256(schema_json).hexdigest()
            })

            data_json = json.dumps(version_info['data'], indent=2).encode('utf-8')
            result_files[f"sampledb_export/objects/{object_info['id']}/versions/{version_info['id']}/data.json"] = data_json
            ro_crate_metadata["@graph"].append({
                "@id": f"./objects/{object_info['id']}/versions/{version_info['id']}/data.json",
                "@type": "File",
                "description": f"Data for Object #{object_info['id']} version #{version_info['id']}",
                "name": "data.json",
                "encodingFormat": "application/json",
                "contentSize": str(len(data_json)),
                "sha256": hashlib.sha256(data_json).hexdigest()
            })

        for comment in object_info['comments']:
            ro_crate_metadata["@graph"].append({
                "@id": f"./objects/{object_info['id']}/comments/{comment['id']}",
                "@type": "Comment",
                "parentItem": {
                    "@id": f"./objects/{object_info['id']}/"
                },
                "author": {"@id": f"./users/{comment['author_id']}"} if comment['author_id'] is not None else None,
                "dateCreated": comment['utc_datetime'],
                "text": comment['content']
            })
            directory_datasets[f"sampledb_export/objects/{object_info['id']}/comments"]['comment'].append({
                "@id": f"./objects/{object_info['id']}/comments/{comment['id']}"
            })

        files_json = json.dumps(object_info['files'], indent=2).encode('utf-8')
        result_files[f"sampledb_export/objects/{object_info['id']}/files.json"] = files_json
        ro_crate_metadata["@graph"].append({
            "@id": f"./objects/{object_info['id']}/files.json",
            "@type": "File",
            "description": f"Data about files for Object #{object_info['id']}",
            "name": "files.json",
            "encodingFormat": "application/json",
            "contentSize": str(len(files_json)),
            "sha256": hashlib.sha256(files_json).hexdigest()
        })

        for file_info in object_info['files']:
            if 'path' in file_info:
                file_name = 'sampledb_export/' + file_info['path']
                file_content = archive_files[file_name]
                relative_file_name = './' + file_info['path']

                charset = None
                if isinstance(file_content, str):
                    file_content = file_content.encode('utf-8')
                    charset = 'UTF-8'

                file_hash = hashlib.sha256(file_content).hexdigest()

                file_extension = os.path.splitext(relative_file_name)[1]
                supported_mime_types = flask.current_app.config['MIME_TYPES']
                if file_extension in flask.current_app.config['MIME_TYPES']:
                    file_type = supported_mime_types[file_extension]
                else:
                    file_type, _ = mimetypes.guess_type(relative_file_name)

                if charset is not None:
                    file_type = f'{file_type}; charset={charset}'

                ro_crate_metadata["@graph"].append({
                    "@id": relative_file_name,
                    "@type": "File",
                    "name": relative_file_name.rsplit('/', 1)[-1],
                    "description": f"File #{file_info['id']} for Object #{object_info['id']}",
                    "author": {
                        "@id": f"./users/{file_info['uploader_id']}"
                    } if file_info['uploader_id'] is not None else None,
                    "dateCreated": file_info['utc_datetime'],
                    "encodingFormat": file_type,
                    "contentSize": str(len(file_content)),
                    "contentUrl": flask.url_for('frontend.object_file', object_id=object_info['id'], file_id=file_info['id'], _external=True),
                    "sha256": file_hash
                })
                result_files[file_name] = file_content

                directory_datasets[f"sampledb_export/objects/{object_info['id']}/files"]["hasPart"].append({
                    "@id": relative_file_name
                })
            else:
                continue

    if not any(user_info['id'] == user_id for user_info in infos['users']):
        user = get_user(user_id)
        infos['users'].append({
            'id': user_id,
            'name': user.name,
            'ordic_id': user.orcid
        })
    for user_info in infos['users']:
        ro_crate_metadata["@graph"].append({
            "@id": f"./users/{user_info['id']}",
            "@type": "Person",
            "name": user_info['name'],
            "url": flask.url_for('frontend.user_profile', user_id=user_info['id'], _external=True)
        })
        if user_info.get('orcid_id'):
            ro_crate_metadata["@graph"][-1]['identifier'] = user_info['orcid_id']

    ro_crate_metadata["@graph"].append({
        "@id": "https://scientific-it-systems.iffgit.fz-juelich.de/SampleDB/",
        "@type": "SoftwareApplication",
        "url": "https://scientific-it-systems.iffgit.fz-juelich.de/SampleDB/",
        "name": "SampleDB",
        "version": version.__version__
    })
    ro_crate_metadata["@graph"].append({
        "@id": "#ro-crate-created",
        "@type": "CreateAction",
        "object": {
            "@id": "./"
        },
        "name": "RO-Crate created",
        "endTime": datetime.date.today().strftime('YYYY-MM-DD'),
        "agent": {"@id": f"./users/{user_id}"},
        "instrument": {
            "@id": "https://scientific-it-systems.iffgit.fz-juelich.de/SampleDB/"
        },
        "actionStatus": {
            "@id": "http://schema.org/CompletedActionStatus"
        }
    })

    result_files['sampledb_export/ro-crate-metadata.json'] = json.dumps(_unpack_single_item_arrays(ro_crate_metadata), indent=2).encode('utf-8')
    result_files['sampledb_export/ro-crate-metadata.json.minisig'] = _sign_ro_crate_metadata(result_files['sampledb_export/ro-crate-metadata.json'])
    result_files['sampledb_export/ro-crate-preview.html'] = _create_preview_html(user_id, list(exported_object_ids), result_files).encode('utf-8')
    return result_files


def _create_preview_html(
        user_id: int,
        object_ids: typing.List[int],
        result_files: typing.Dict[str, bytes]
) -> str:
    from ..frontend.markdown_images import IMAGE_FORMATS
    from ..frontend.pdfexport import create_html_for_pdfexport

    base_url = flask.url_for('.index', _external=True)

    def url_mapper(url: str, exported_files: typing.Dict[typing.Tuple[int, int], File]) -> str:
        if url.startswith(base_url + 'markdown_images/'):
            file_name = url[len(base_url + 'markdown_images/'):]
            image_data = markdown_images.get_markdown_image(file_name, user_id)
            if image_data is None:
                return ''
            file_extension = os.path.splitext(file_name)[1].lower()
            if file_extension not in IMAGE_FORMATS:
                return ''
            result_files['sampledb_export/ro-crate-preview_files/markdown_images/' + file_name] = image_data
            return 'ro-crate-preview_files/markdown_images/' + file_name
        if url.startswith(base_url + 'objects/') and '/files/' in url[len(base_url + 'objects/'):]:
            found_result_file = [
                file_path[len('sampledb_export/'):]
                for file_path in result_files
                if file_path.startswith('sampledb_export/' + url[len(base_url):] + '/')
            ]
            if found_result_file:
                return found_result_file[0]
            object_id_file_id = url[len(base_url + 'objects/'):]
            url = ''
            try:
                object_id_str, file_id_str = object_id_file_id.split('/files/')
                object_id = int(object_id_str)
                file_id = int(file_id_str)
                file = exported_files[(object_id, file_id)]
                if file.storage == 'database' and not file.is_hidden:
                    for file_extension, mime_type in IMAGE_FORMATS.items():
                        if file.original_file_name.lower().endswith(file_extension):
                            image_data = file.open(read_only=True).read()
                            return 'data:' + mime_type + ';base64,' + base64.b64encode(image_data).decode('utf-8')
            except Exception:
                pass
        # only allow Data URLs and URLs via http or https
        if not (url.startswith('data:') or urllib.parse.urlparse(url).scheme in ('http', 'https')):
            return ''
        return url

    return create_html_for_pdfexport(
        user_id=user_id,
        object_ids=object_ids,
        url_mapper=url_mapper,
        include_logo=False
    )


def _sign_ro_crate_metadata(
    data: bytes
) -> bytes:
    kp = minisign_keys.get_current_key_pair()
    secret_key = minisign.SecretKey.from_bytes(kp.sk_bytes)
    sig = secret_key.sign(data, trusted_comment=url_for('frontend.key_list_json', _external=True))
    return bytes(sig)


def _convert_metadata_to_property_values(
        object_id: int,
        data: typing.Dict[str, typing.Any],
        schema: typing.Dict[str, typing.Any],
        exported_object_ids: typing.Dict[int, str],
        exported_user_ids: typing.Dict[int, str],
        exported_file_ids: typing.Dict[int, str],
        property_id_prefix: str
) -> typing.List[typing.Dict[str, typing.Any]]:
    property_values = []
    for property_data, _property_path, full_property_path in flatten_metadata(data):
        property_type = property_data.get('_type')
        if not property_type:
            continue
        property_id = '.'.join(str(path_element) for path_element in full_property_path)
        property_value: typing.Dict[str, typing.Any] = {
            "@type": "PropertyValue",
            "@id": property_id_prefix + property_id,
            "propertyID": property_id,
            "name": get_title_for_property(full_property_path, schema)
        }
        if property_type == 'text':
            property_values.append({
                "value": get_translated_text(property_data['text'], 'en'),
                **property_value
            })
        elif property_type == 'bool':
            property_values.append({
                "value": property_data['value'],
                **property_value
            })
        elif property_type == 'datetime':
            property_values.append({
                "value": datetime.datetime.strptime(property_data['utc_datetime'], '%Y-%m-%d %H:%M:%S').replace(tzinfo=None).isoformat(timespec='microseconds'),
                **property_value
            })
        elif property_type == 'quantity':
            quantity = Quantity.from_json(property_data)
            un_cefact_code = get_un_cefact_code_for_unit(quantity.units)
            if un_cefact_code is not None:
                property_value["unitCode"] = un_cefact_code
            property_values.append({
                "value": quantity.magnitude,
                "unitText": quantity.units,
                **property_value
            })
        elif property_type in ('sample', 'measurement', 'object_reference'):
            property_values.append({
                "value": exported_object_ids.get(property_data['object_id'], flask.url_for('frontend.object', object_id=property_data['object_id'], _external=True)),
                **property_value
            })
        elif property_type == 'user':
            property_values.append({
                "value": exported_user_ids.get(property_data['user_id'], flask.url_for('frontend.user_profile', user_id=property_data['user_id'], _external=True)),
                **property_value
            })
        elif property_type == 'file':
            property_values.append({
                "value": exported_file_ids.get(property_data['file_id'], flask.url_for('frontend.object_file', object_id=object_id, file_id=property_data['file_id'], _external=True)),
                **property_value
            })
        elif property_type == 'hazards':
            hazard_texts = {1: 'Explosive', 2: 'Flammable', 3: 'Oxidizing', 4: 'Compressed Gas', 5: 'Corrosive', 6: 'Toxic', 7: 'Harmful', 8: 'Health Hazard', 9: 'Environmental Hazard'}
            property_values.append({
                "value": ", ".join(f"{hazard_texts[hazard]}" for hazard in sorted(property_data['hazards'])),
                **property_value
            })
    return property_values



# coding: utf-8
"""
Script for generating an UUID to be used as database identifier in a federation.

Usage: sampledb generate_uuid
"""

import sys
import typing
import uuid


def main(arguments: typing.List[str]) -> None:
    if len(arguments) != 0:
        print(__doc__)
        sys.exit(1)
    print(uuid.uuid4())



import typing

from .. import db
from ..models import ComponentAuthentication, ComponentAuthenticationType, OwnComponentAuthentication
from .errors import NoAuthenticationMethodError, InvalidTokenError, AuthenticationMethodDoesNotExistError, \
    TokenExistsError
from .authentication import _hash_password, _validate_password_hash
from .components import Component


def get_own_authentication(
        component_id: int,
        type: ComponentAuthenticationType = ComponentAuthenticationType.TOKEN
) -> OwnComponentAuthentication:
    auth: typing.Optional[OwnComponentAuthentication] = OwnComponentAuthentication.query.filter_by(component_id=component_id, type=type).first()
    if auth is None:
        raise NoAuthenticationMethodError()
    return auth


def own_token_exists(
        component_id: int,
        token: str
) -> bool:
    authentication_methods = OwnComponentAuthentication.query.filter(
        db.and_(OwnComponentAuthentication.login['token'].astext == token,
                db.and_(OwnComponentAuthentication.type == ComponentAuthenticationType.TOKEN,
                        OwnComponentAuthentication.component_id == component_id))
    ).first()
    return authentication_methods is not None


def add_token_authentication(
        component_id: int,
        token: str,
        description: str
) -> None:
    if len(token) != 64:
        raise InvalidTokenError()
    token = token.lower()
    # split into a short part (8 hex digits / 4 bytes) for identification and a long part for authentication
    login, password = token[:8], token[8:]
    authentication = ComponentAuthentication(
        login={
            'login': login,
            'bcrypt_hash': _hash_password(password),
            'description': description
        },
        authentication_type=ComponentAuthenticationType.TOKEN,
        component_id=component_id
    )
    db.session.add(authentication)
    db.session.commit()


def add_own_token_authentication(
        component_id: int,
        token: str,
        description: str
) -> None:
    token = token.lower().strip()
    if len(token) != 64:
        raise InvalidTokenError()
    if own_token_exists(component_id, token):
        raise TokenExistsError()
    authentication = OwnComponentAuthentication(
        login={
            'token': token,
            'description': description
        },
        authentication_type=ComponentAuthenticationType.TOKEN,
        component_id=component_id
    )
    db.session.add(authentication)
    db.session.commit()


def login_via_component_token(
        component_token: str
) -> typing.Optional[Component]:
    """
    Authenticate a component using an API token.

    :param component_token: the API token to use during authentication
    :return: the component or None
    """
    # convert to lower case to enforce case insensitivity
    component_token = component_token.lower().strip()
    login, password = component_token[:8], component_token[8:]
    authentication_methods = ComponentAuthentication.query.filter(
        db.and_(ComponentAuthentication.login['login'].astext == login,
                ComponentAuthentication.type == ComponentAuthenticationType.TOKEN)
    ).all()

    for authentication_method in authentication_methods:
        if _validate_password_hash(password, authentication_method.login['bcrypt_hash']):
            if authentication_method.component:
                return Component.from_database(authentication_method.component)
    return None


def remove_component_authentication_method(
        authentication_method_id: int
) -> None:
    """
    Remove a component authentication method.

    :param authentication_method_id: the ID of an existing authentication method
    :raise error.AuthenticationMethodDoesNotExistError: when no component
        authentication method with the given ID exists
    """
    authentication_method = ComponentAuthentication.query.filter(ComponentAuthentication.id == authentication_method_id).first()
    if authentication_method is None:
        raise AuthenticationMethodDoesNotExistError()

    db.session.delete(authentication_method)
    db.session.commit()


def remove_own_component_authentication_method(
        authentication_method_id: int
) -> None:
    """
    Remove a component authentication method.

    :param authentication_method_id: the ID of an existing authentication method
    :raise error.AuthenticationMethodDoesNotExistError: when no component
        authentication method with the given ID exists
    """
    authentication_method = OwnComponentAuthentication.query.filter(OwnComponentAuthentication.id == authentication_method_id).first()
    if authentication_method is None:
        raise AuthenticationMethodDoesNotExistError()

    db.session.delete(authentication_method)
    db.session.commit()



# coding: utf-8
"""

"""
from base64 import b64decode
import typing

import flask
import flask_login
from flask_babel import _
from flask_wtf import FlaskForm
from wtforms import FileField, FieldList, FormField, SelectField, StringField, BooleanField
from wtforms.validators import InputRequired, ValidationError

from . import frontend
from .federation_forms import ModifyELNIdentityForm
from .utils import check_current_user_is_not_readonly
from .. import logic
from ..utils import FlaskResponseT
from ..logic.errors import UserDoesNotExistError, ELNUserInFederatedIdentityError
from ..models.users import FederatedIdentity


class UploadELNFileForm(FlaskForm):
    eln_file = FileField(validators=[InputRequired()])


class ImportELNFileObjectForm(FlaskForm):
    action_type_id = SelectField(coerce=int, validators=[InputRequired()])


class ImportELNFileForm(FlaskForm):
    object_forms = FieldList(FormField(ImportELNFileObjectForm))
    federated_identities = StringField()


class DeleteELNImportForm(FlaskForm):
    submit = StringField(validators=[InputRequired()])

    def validate_submit(form, field: StringField) -> None:
        if not field.data == 'delete':
            raise ValidationError("")


@frontend.route('/eln_imports', methods=['POST', 'GET'])
@flask_login.login_required
def upload_eln_import() -> FlaskResponseT:
    if not flask.current_app.config['ENABLE_ELN_FILE_IMPORT']:
        return flask.abort(404)
    check_current_user_is_not_readonly()

    logic.eln_import.remove_expired_eln_imports()

    upload_eln_file_form = UploadELNFileForm()

    if upload_eln_file_form.validate_on_submit():
        eln_file = upload_eln_file_form.eln_file.data
        zip_bytes = eln_file.read()
        # TODO: check that zip_bytes belong to ELN file
        eln_import_id = logic.eln_import.create_eln_import(
            user_id=flask_login.current_user.id,
            file_name=eln_file.filename,
            zip_bytes=zip_bytes
        ).id
        return flask.redirect(flask.url_for('.eln_import', eln_import_id=eln_import_id))

    pending_eln_imports = logic.eln_import.get_pending_eln_imports(flask_login.current_user.id)
    delete_eln_import_forms = [
        DeleteELNImportForm()
        for _ in pending_eln_imports
    ]

    return flask.render_template(
        'eln_files/upload_eln_file.html',
        upload_eln_file_form=upload_eln_file_form,
        pending_eln_imports=pending_eln_imports,
        delete_eln_import_forms=delete_eln_import_forms,
    )


@frontend.route('/eln_imports/<int:eln_import_id>', methods=['POST', 'GET'])
@flask_login.login_required
def eln_import(eln_import_id: int) -> FlaskResponseT:
    logic.eln_import.remove_expired_eln_imports()

    try:
        eln_import = logic.eln_import.get_eln_import(eln_import_id)
    except logic.errors.ELNImportDoesNotExistError:
        return flask.abort(404)
    if not eln_import.imported and eln_import.user_id != flask_login.current_user.id and not flask_login.current_user.has_admin_permissions:
        return flask.abort(403)
    if eln_import.imported:
        imported_users = logic.eln_import.get_eln_import_users(eln_import_id)
        fed_identities: dict[int, list[tuple[typing.Optional[ModifyELNIdentityForm], FederatedIdentity]]] = {}

        for eln_user in imported_users:
            identity = logic.users.get_federated_identity_by_eln_user(eln_user.id)
            if identity:
                identities = fed_identities.get(identity.user.id, [])
                form = ModifyELNIdentityForm(eln_user_id=identity.local_fed_id) if identity.user_id == flask_login.current_user.id else None
                identities.append((form, identity))
                fed_identities[identity.user.id] = identities

        return flask.render_template(
            'eln_files/view_eln_file.html',
            eln_import=eln_import,
            imported_objects=logic.eln_import.get_eln_import_objects(eln_import_id),
            imported_users=logic.eln_import.get_eln_import_users(eln_import_id),
            fed_identities=fed_identities
        )
    else:
        if not flask.current_app.config['ENABLE_ELN_FILE_IMPORT']:
            return flask.abort(404)
        check_current_user_is_not_readonly()

        delete_eln_import_form = DeleteELNImportForm()
        if delete_eln_import_form.validate_on_submit():
            logic.eln_import.delete_eln_import(eln_import_id)
            flask.flash(_("Successfully deleted .eln file."), 'success')
            return flask.redirect(flask.url_for('.upload_eln_import'))

        try:
            parsed_eln_file_data = logic.eln_import.parse_eln_file(
                eln_import_id=eln_import_id
            )
        except logic.errors.InvalidELNFileError as e:
            message = str(e)
            logic.eln_import.mark_eln_import_invalid(eln_import_id, message)
            flask.flash(_("Failed to parse .eln file: %(message)s", message=message), 'error')
            return flask.redirect(flask.url_for('.upload_eln_import'))

        valid_action_types = [
            (action_type.id, action_type)
            for action_type in logic.action_types.get_action_types()
            if action_type.component_id is None and not action_type.disable_create_objects
        ]
        import_eln_file_form = ImportELNFileForm(
            object_forms=[ImportELNFileObjectForm()] * len(parsed_eln_file_data.objects),
            attach_federated_identity=[BooleanField(name="x")] * len(parsed_eln_file_data.users)
        )
        for object_data, object_form in zip(parsed_eln_file_data.objects, import_eln_file_form.object_forms.entries):
            object_form.action_type_id.choices = valid_action_types
            if not import_eln_file_form.is_submitted() and object_data.type_id is not None:
                object_form.action_type_id.data = object_data.type_id

        if import_eln_file_form.validate_on_submit():
            imported_object_ids, users_by_id, errors = logic.eln_import.import_eln_file(
                eln_import_id=eln_import_id,
                action_type_ids=[
                    object_form.action_type_id.data
                    for object_form in import_eln_file_form.object_forms.entries
                ]
            )

            if errors:
                if imported_object_ids:
                    flask.flash(
                        _('%(num_imported_objects)s objects were imported, but the following errors occurred:', num_imported_objects=len(imported_object_ids)) + ' ' + ' '.join(errors),
                        'error'
                    )
                else:
                    flask.flash(
                        _('The following errors occurred while trying to import objects:', num_imported_objects=len(imported_object_ids)) + ' ' + ' '.join(errors),
                        'error'
                    )
                    return flask.redirect(flask.url_for('.eln_import', eln_import_id=eln_import_id))
            else:
                if len(imported_object_ids) == 1:
                    flask.flash(
                        _('Successfully imported an object.'),
                        'success'
                    )
                else:
                    flask.flash(
                        _('Successfully imported %(num_imported_objects)s objects.', num_imported_objects=len(imported_object_ids)),
                        'success'
                    )

            failed_identities: list[str] = []
            for encoded_user_id in import_eln_file_form.federated_identities.data.split(","):
                eln_user_id = b64decode(encoded_user_id).decode("utf-8")
                database_user = users_by_id.get(eln_user_id, None)
                if database_user:
                    try:
                        logic.users.create_eln_federated_identity(user_id=flask_login.current_user.id, eln_user_id=database_user.id)
                    except (UserDoesNotExistError, ELNUserInFederatedIdentityError):
                        failed_identities.append(eln_user_id)

            if failed_identities:
                flask.flash(_("An error occured: Failed to create federated identity."), "error")

            if len(imported_object_ids) == 1:
                return flask.redirect(flask.url_for('.object', object_id=imported_object_ids[0]))
            else:
                return flask.redirect(flask.url_for('.objects', ids=','.join(str(object_id) for object_id in imported_object_ids)))
        return flask.render_template(
            'eln_files/import_eln_file.html',
            import_eln_file_form=import_eln_file_form,
            parsed_eln_file_data=parsed_eln_file_data,
        )



# coding: utf-8
"""
RESTful API for SampleDB
"""
import typing

import flask

from .authentication import multi_auth
from ..utils import Resource, ResponseData
from ...logic import errors, groups


def group_to_json(group: groups.Group) -> typing.Dict[str, typing.Any]:
    return {
        'group_id': group.id,
        'name': group.name,
        'description': group.description,
        'member_users': [{
            'user_id': user_id,
            'href': flask.url_for('api.user', user_id=user_id, _external=True),
        } for user_id in sorted(groups.get_group_member_ids(group_id=group.id))]
    }


class Groups(Resource):
    @multi_auth.login_required
    def get(self) -> ResponseData:
        return [
            group_to_json(group)
            for group in sorted(groups.get_groups(), key=lambda group: group.id)
        ]


class Group(Resource):
    @multi_auth.login_required
    def get(self, group_id: int) -> ResponseData:
        try:
            group = groups.get_group(group_id=group_id)
        except errors.GroupDoesNotExistError:
            return {
                "message": f"group {group_id} does not exist"
            }, 404
        return group_to_json(group)



# coding: utf-8
"""
Add discoverable column to components table.
"""

import flask_sqlalchemy

from .utils import table_has_column


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    if table_has_column('components', 'discoverable'):
        return False

    # Perform migration
    db.session.execute(db.text("""
        ALTER TABLE components
            ADD discoverable BOOLEAN NOT NULL DEFAULT true
    """))
    return True



import typing

from . import users
from .permissions import ResourcePermissions
from .errors import InvalidDefaultPermissionsError
from ..models import Permissions, DefaultUserPermissions, DefaultGroupPermissions, DefaultProjectPermissions, AllUserDefaultPermissions

default_permissions = ResourcePermissions(
    resource_id_name='creator_id',
    all_user_permissions_table=AllUserDefaultPermissions,
    anonymous_user_permissions_table=None,
    user_permissions_table=DefaultUserPermissions,
    group_permissions_table=DefaultGroupPermissions,
    project_permissions_table=DefaultProjectPermissions,
    check_resource_exists=lambda resource_id: users.check_user_exists(user_id=resource_id)  # type: ignore
)


def get_default_permissions_for_users(creator_id: int) -> typing.Dict[int, Permissions]:
    # the creator always has GRANT as default permissions
    additional_permissions = {
        creator_id: Permissions.GRANT
    }
    return default_permissions.get_permissions_for_users(
        resource_id=creator_id,
        additional_permissions=additional_permissions
    )


def set_default_permissions_for_user(creator_id: int, user_id: int, permissions: Permissions) -> None:
    if user_id == creator_id:
        if permissions == Permissions.GRANT:
            return None
        else:
            raise InvalidDefaultPermissionsError("creator will always get GRANT permissions")
    return default_permissions.set_permissions_for_user(
        resource_id=creator_id,
        user_id=user_id,
        permissions=permissions
    )


def get_default_permissions_for_groups(creator_id: int) -> typing.Dict[int, Permissions]:
    return default_permissions.get_permissions_for_groups(resource_id=creator_id)


def set_default_permissions_for_group(creator_id: int, group_id: int, permissions: Permissions) -> None:
    default_permissions.set_permissions_for_group(
        resource_id=creator_id,
        group_id=group_id,
        permissions=permissions
    )


def get_default_permissions_for_projects(creator_id: int) -> typing.Dict[int, Permissions]:
    return default_permissions.get_permissions_for_projects(resource_id=creator_id)


def set_default_permissions_for_project(creator_id: int, project_id: int, permissions: Permissions) -> None:
    default_permissions.set_permissions_for_project(
        resource_id=creator_id,
        project_id=project_id,
        permissions=permissions
    )


def get_default_permissions_for_all_users(creator_id: int) -> Permissions:
    return default_permissions.get_permissions_for_all_users(resource_id=creator_id)


def set_default_permissions_for_all_users(creator_id: int, permissions: Permissions) -> None:
    return default_permissions.set_permissions_for_all_users(resource_id=creator_id, permissions=permissions)


def get_default_permissions_for_anonymous_users(creator_id: int) -> Permissions:
    return default_permissions.get_permissions_for_anonymous_users(resource_id=creator_id)



# coding: utf-8
"""
Logic package for sampledb schemas

The structure of sample and measurement information is defined using schemas,
JSON-like objects similar to JSON schema. Instead of the primitive types used
in JSON schema, sampledb schemas can use the following types:

- bool
- text
- quantity
- datetime
- sample
"""

from . import utils, templates, data_diffs
from .convert_to_schema import convert_to_schema, is_converting_to_schema_necessary
from .copy_data import copy_data
from .data_diffs import apply_diff, calculate_diff
from .generate_placeholder import generate_placeholder, get_default_data
from .validate_schema import validate_schema
from .validate import validate


__all__ = [
    'apply_diff',
    'convert_to_schema',
    'copy_data',
    'calculate_diff',
    'data_diffs',
    'generate_placeholder',
    'get_default_data',
    'is_converting_to_schema_necessary',
    'templates',
    'utils',
    'validate_schema',
    'validate'
]



import dataclasses
import datetime
import typing
import flask

from ..models import temporary_files
from .. import db
from .files import create_database_file


@dataclasses.dataclass(frozen=True)
class TemporaryFile:
    """
    This class provides an immutable wrapper around models.temporary_files.TemporaryFile.
    """
    id: int
    context_id: str
    file_name: str
    user_id: int
    utc_datetime: datetime.datetime
    binary_data: bytes

    @classmethod
    def from_database(cls, temporary_file: temporary_files.TemporaryFile) -> 'TemporaryFile':
        return TemporaryFile(
            id=temporary_file.id,
            context_id=temporary_file.context_id,
            file_name=temporary_file.file_name,
            user_id=temporary_file.user_id,
            utc_datetime=temporary_file.utc_datetime,
            binary_data=temporary_file.binary_data
        )


def get_files_for_context_id(context_id: str) -> typing.List[TemporaryFile]:
    """
    Get all temporary files saved in a given context.

    :param context_id: the context ID for which to return all temporary files
    """
    return [
        TemporaryFile.from_database(temporary_file)
        for temporary_file in temporary_files.TemporaryFile.query.filter_by(context_id=context_id).order_by(db.asc(temporary_files.TemporaryFile.id)).all()
    ]


def delete_expired_temporary_files() -> None:
    """
    Delete all expired temporary files.
    """
    time_limit = flask.current_app.config['TEMPORARY_FILE_TIME_LIMIT']
    db.session.query(temporary_files.TemporaryFile).filter(
        temporary_files.TemporaryFile.utc_datetime < datetime.datetime.now(datetime.timezone.utc) - datetime.timedelta(seconds=time_limit)
    ).delete()
    db.session.commit()


def delete_temporary_files(
        context_id: str
) -> None:
    """
    Delete all temporary files saved in a given context.

    :param context_id: the context ID for which to delete all temporary files
    """
    db.session.query(temporary_files.TemporaryFile).filter(
        temporary_files.TemporaryFile.context_id == context_id
    ).delete()
    db.session.commit()


def create_temporary_file(
        context_id: str,
        file_name: str,
        user_id: int,
        binary_data: bytes
) -> TemporaryFile:
    """
    Create a temporary file.

    :param context_id: the context ID for the temporary file
    :param file_name: the name for the file
    :param user_id: the ID of the user who uploaded the file
    :param binary_data: the data for the file
    :return: the created file
    """
    temporary_file = temporary_files.TemporaryFile(
        context_id=context_id,
        file_name=file_name,
        user_id=user_id,
        binary_data=binary_data,
        utc_datetime=datetime.datetime.now(datetime.timezone.utc)
    )
    db.session.add(temporary_file)
    db.session.commit()
    return TemporaryFile.from_database(temporary_file)


def copy_temporary_files(
        file_ids: typing.Sequence[int],
        context_id: str,
        user_id: int,
        object_ids: typing.Sequence[int]
) -> typing.Dict[int, int]:
    """
    Copy a list of temporary files to one or more objects.

    :param file_ids: the IDs of existing temporary files
    :param context_id: the context ID for the temporary files
    :param user_id: the ID of the user copying the files
    :param object_ids: the IDs of objects to copy the files to
    :return: a mapping between temporary and permanent file IDs
    """
    temporary_file_id_map = {}
    temporary_files = get_files_for_context_id(context_id=context_id)
    for file in temporary_files:
        if file.id in file_ids:
            def save_content(stream: typing.BinaryIO, file: TemporaryFile = file) -> None:
                stream.write(file.binary_data)
            for object_id in object_ids:
                temporary_file_id_map[file.id] = create_database_file(
                    object_id=object_id,
                    user_id=user_id,
                    file_name=file.file_name,
                    save_content=save_content
                ).id
    return temporary_file_id_map


def replace_file_reference_ids(
        data: typing.Any,
        temporary_file_id_map: typing.Dict[int, int]
) -> None:
    if isinstance(data, dict):
        if '_type' in data:
            if data['_type'] == 'file':
                if -data['file_id'] in temporary_file_id_map:
                    data['file_id'] = temporary_file_id_map[-data['file_id']]
        else:
            for property_data in data.values():
                replace_file_reference_ids(property_data, temporary_file_id_map)
    if isinstance(data, list):
        for item_data in data:
            replace_file_reference_ids(item_data, temporary_file_id_map)


def get_referenced_temporary_file_ids(
        data: typing.Any
) -> typing.Set[int]:
    refrenced_file_ids = set()
    if isinstance(data, dict):
        if '_type' in data:
            if data['_type'] == 'file' and data['file_id'] < 0:
                refrenced_file_ids.add(-data['file_id'])
        else:
            for property_data in data.values():
                refrenced_file_ids = refrenced_file_ids.union(get_referenced_temporary_file_ids(property_data))
    if isinstance(data, list):
        for item_data in data:
            refrenced_file_ids = refrenced_file_ids.union(get_referenced_temporary_file_ids(item_data))
    return refrenced_file_ids
